# -*- coding: utf-8 -*-
# 
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Base classes for modeling domain objects that wrap vertices and edges.

"""
import six  # Python 3
import inspect
import types
from collections import Callable

from bulbs.property import Property
from bulbs.element import Element, Vertex, VertexProxy, Edge, EdgeProxy, \
    coerce_vertices, build_data
from bulbs.utils import initialize_element, get_logger


# Model Modes
NORMAL = 1
STRICT = 2

log = get_logger(__name__)


class ModelMeta(type):
    """Metaclass used to set database Property definitions on Models."""

    def __init__(cls, name, base, namespace):
        """Store Property instance definitions on the class as a dictionary.""" 

        # Get inherited Properties
        cls._properties = cls._get_initial_properties()
        
        # Add new Properties
        cls._register_properties(namespace)

    def _get_initial_properties(cls):
        """
        Get Properties defined in the parent and inherit them.

        :rtype: dict

        """
        try: 
            parent_properties = getattr(cls, '_properties')
            properties = parent_properties.copy() 
        except:
            # Set it to an empty dict if the Model doesn't have a parent Model. 
            properties = {}
        return properties
            
    def _register_properties(cls, namespace):
        """
        Loop through the class namespace looking for Property instances.

        :param namespace: Class namespace
        :type namespace: dict

        :rtype: None

        """

        # e.g. age = Integer()
        for key in namespace: # Python 3
            value = namespace[key]

            assert key not in cls._properties, \
                "Can't redefine Property '%s'" % key

            if isinstance(value, Property):
                property_instance = value  # for clarity
                cls._properties[key] = property_instance
                cls._set_property_name(key,property_instance)
                cls._initialize_property(key,property_instance)
                # not doing this b/c some Properties are calculated at savetime
                #delattr(cls, key) 
                            
    def _set_property_name(cls, key, property_instance):
        """
        Set Property name to attribute key unless explicitly set via kwd param.

        :param key: Class attribute key
        :type key: str

        :param property_instance: Property instance
        :type property_instance bulbs.property.Property

        :rtype None

        """
        if property_instance.name is None:
            property_instance.name = key

    def _initialize_property(cls, key, property_instance):
        """
        Set the Model class attribute based on the Property definition.

        :param key: Class attribute key
        :type key: str

        :param property_instance: Property instance
        :type property_instance bulbs.property.Property

        """
        if property_instance.fget:
            # TODO: make this configurable
            # this is a calculated property (should it persist?)
            # wrapped fset and fdel in str() to make the default None not 
            # error on getattr
            fget = getattr(cls, property_instance.fget)
            # TODO: implement fset and fdel (maybe)
            #fset = getattr(cls, str(property_instance.fset), None)
            #fdel = getattr(cls, str(property_instance.fdel), None)
            fset = None
            fdel = None
            property_value = property(fget, fset, fdel)
        else:
            property_value = None
        setattr(cls, key, property_value)


class Model(six.with_metaclass(ModelMeta, object)):  # Python 3
    """Abstract base class for Node and Relationship container classes."""
    

    #: The mode for saving attributes to the database. 
    #: If set to STRICT, only defined Properties are saved.
    #: If set to NORMAL, all attributes are saved. 
    #: Defaults to NORMAL. 
    __mode__ = NORMAL
    
    #: A dict containing the database Property instances.
    _properties = None
    

    def __setattr__(self, key, value):
        """
        Set model attributes, possibly coercing database Properties to the 
        defined types.

        :param key: Attribute key
        :type key: str

        :param value: Attribute value
        :type value: object
        
        :rtype: None

        """
        if key in self._properties:
            self._set_database_property(key, value)
        else:
            # If _mode = STRICT, set an instance var, which isn't saved to DB.
            # If _mode = NORMAL, store in self._data, which is saved to DB
            self._set_normal_attribute(key, value)

    def _set_database_property(self, key, value):
        """
        Set Property attributes after coercing them into the defined types.

        :param key: Attribute key
        :type key: str

        :param value: Attribute value
        :type value: object
        
        :rtype: None

        """
        # we want Model Properties to be set be set as actual attributes
        # because they can be real Python propertes or calculated values,
        # which are calcualted/set upon each save().
        # Don't set calculated (fget) properties; they're calculated at save.
        if not self._is_calculated_property(key):
            value = self._coerce_property_value(key, value)
            object.__setattr__(self, key, value)

    def _set_normal_attribute(self, key, value):
        """
        Set normal/non-database Property attributes, depending on the __mode__.

        :param key: Attribute key
        :type key: str

        :param value: Attribute value
        :type value: object
        
        :rtype: None

        """
        if self.__mode__ == STRICT:
            # Set as a Python attribute, which won't be saved to the database.
            object.__setattr__(self, key, value)
        else:
            # Store the attribute in self._data, which are saved to database.
            Element.__setattr__(self, key, value)        

    def _is_calculated_property(self, key):
        """
        Returns True if the Property is a cacluated property, i.e. has fget set.

        :param key: Attribute key
        :type key: str

        :rtype: bool

        """
        # TODO: fget works, but fset, fdel have not been tested
        property_instance = self._properties[key]
        return (property_instance.fget is not None)

    def _coerce_property_value(self, key, value):
        """
        Coerce database Property value into its defined type.

        :param key: Attribute key
        :type key: str

        :param value: Attribute value
        :type value: object
        
        :rtype: object

        """
        if value is not None:
            property_instance = self._properties[key]
            value = property_instance.coerce(key, value)
        return value

    def _set_property_defaults(self):
        """
        Set the default values for all the database Properties.

        :rtype: None

        """
        for key in self._properties:
            default_value = self._get_property_default(key)     
            setattr(self, key, default_value)
            
    def _get_property_default(self, key):
        """
        Coerce database Property value into its defined type.

        :param key: Attribute key
        :type key: str

        :rtype: object

        """
        # TODO: make this work for model methods?
        # The value entered could be a scalar or a function name
        # Should we defer the call until all properties are set, 
        # or only for calculated properties?
        property_instance = self._properties[key]
        default_value = property_instance.default
        if isinstance(default_value, Callable):
            default_value = default_value()
        return default_value

    def _set_keyword_attributes(self, _data, kwds):
        """
        Sets Python attributes using the _data and keywords passed in by user.

        :param _data: Data that was passed in via a dict.
        :type _data: dict

        :param kwds: Data that was passed in via name/value pairs.
        :type kwds: dict

        :rtype: None

        """
        # NOTE: keys may have been passed in that are not defined as Properties
        data = build_data(_data, kwds)
        for key in data:    # Python 3
            value = data[key]  
            # Notice that __setattr__ is overloaded
            setattr(self, key, value)

    def _set_property_data(self):
        """
        Sets Property data after it is retrieved from the DB.

        :rtype: None

        .. note:: Sets the value to None if it's an invalid type.

        """
        type_system = self._client.type_system
        for key in self._properties:   # Python 3
            
            # Don't set calculted property values, i.e. those with fset defined.
            if self._is_calculated_property(key): continue

            property_instance = self._properties[key]
            #name = property_instance.name
            value = self._data.get(key, None)
            value = property_instance.convert_to_python(type_system, key, value)

            # TODO: Maybe need to wrap this in try/catch too.
            # Notice that __setattr__ is overloaded. No need to coerce it twice.
            object.__setattr__(self, key, value)
            
    def _get_property_data(self):
        """
        Returns validated Property data, ready to be saved in the DB.

        :rtype: dict

        """
        # If __mode__ is STRICT, data set to empty; otherwise set to self._data
        data = self._get_initial_data()

        type_var = self._client.config.type_var
        type_system = self._client.type_system

        if hasattr(self, type_var):
            # Add element_type to the database properties to be saved;
            # but don't worry about "label", it's always saved on the edge.
            data[type_var] = object.__getattribute__(self, type_var)

        # Convert database Property values to their database types.
        for key in self._properties:  # Python 3
            property_instance = self._properties[key]
            value = self._get_property_value(key)
            property_instance.validate(key, value)
            #name = property_instance.name
            db_value = property_instance.convert_to_db(type_system, key, value)
            data[key] = db_value

        return data

    def _get_initial_data(self):
        """
        Returns empty dict if __mode__ is set to STRICT, otherwise self._data.
        
        :rtype: dict

        """
        data = {} if self.__mode__ == STRICT else self._data.copy()
        return data

    def _get_property_value(self, key):
        """
        Returns the value of a Property, calculated via a function if needed.

        :param key: Attribute key
        :type key: str

        :rtype: object

        """
        # Notice that __getattr__ is overloaded in Element.
        value = object.__getattribute__(self, key)
        if isinstance(value, Callable):
            return value()
        return value

    def get_bundle(self, _data=None, **kwds):
        """
        Returns a tuple contaning the property data, index name, and index keys.

        :param _data: Data that was passed in via a dict.
        :type _data: dict

        :param kwds: Data that was passed in via name/value pairs.
        :type kwds: dict

        :rtype: tuple

        """
        self._set_property_defaults()   
        self._set_keyword_attributes(_data, kwds)
        data = self._get_property_data()
        index_name = self.get_index_name(self._client.config)
        keys = self.get_index_keys()
        return data, index_name, keys

    def get_index_keys(self):
        """
        Returns Property keys to index in DB. Defaults to None (index all keys).

        :rtype: list or None

        """
        # TODO: Derive this from Property definitions.
        return None

    def get_property_keys(self):
        """
        Returns a list of all the Property keys.

        :rtype: list

        """
        return self._properties.keys()

    def data(self):
        """
        Returns a the element's property data.

        :rtype: dict

        """
        data = dict()
        if self.__mode__ == NORMAL:
            data = self._data

        for key in self._properties: 
            # TODO: make this work for calculated values.
            # Calculated props shouldn't be stored, but components should be.
            data[key] = object.__getattribute__(self, key)
        return data

    def map(self):
        """
        Deprecated. Returns the element's property data.

        :rtype: dict

        """
        log.debug("This is deprecated; use data() instead.")
        return self.data()


class Node(Model, Vertex):
    """ 
    Abstract base class used for creating a Vertex Model.
 
    It's used to create classes that model domain objects, and it's not meant 
    to be used directly. To use it, create a subclass specific to the type of 
    data you are storing. 

    Example model declaration::

        # people.py

        from bulbs.model import Node
        from bulbs.property import String, Integer

        class Person(Node):
            element_type = "person"
            
            name = String(nullable=False)
            age = Integer()

    Example usage::

        >>> from people import Person
        >>> from bulbs.neo4jserver import Graph
        >>> g = Graph()

        # Add a "people" proxy to the Graph object for the Person model:
        >>> g.add_proxy("people", Person)

        # Use it to create a Person node, which also saves it in the database:
        >>> james = g.people.create(name="James")
        >>> james.eid
        3
        >>> james.name
        'James'

        # Get the node (again) from the database by its element ID:
        >>> james = g.people.get(james.eid)

        # Update the node and save it in the database:
        >>> james.age = 34
        >>> james.save()

        # Lookup people using the Person model's primary index:
        >>> nodes = g.people.index.lookup(name="James")
        
    """
    #: The mode for saving attributes to the database. 
    #: If set to STRICT, only defined Properties are saved.
    #: If set to NORMAL, all attributes are saved. 
    #: Defaults to NORMAL. 
    __mode__ = NORMAL
    
    #: A dict containing the database Property instances.
    _properties = None

    element_type = None

    @classmethod
    def get_element_type(cls, config):
        """
        Returns the element type.

        :param config: Config object.
        :type config: bulbs.config.Config

        :rtype: str

        """
        element_type = getattr(cls, config.type_var)
        return element_type

    @classmethod
    def get_element_key(cls, config):
        """
        Returns the element key.

        :param config: Config object.
        :type config: bulbs.config.Config

        :rtype: str

        """
        return cls.get_element_type(config)

    @classmethod 
    def get_index_name(cls, config):
        """
        Returns the index name. 

        :param config: Config object.
        :type config: bulbs.config.Config

        :rtype: str

        """
        return cls.get_element_type(config)

    @classmethod 
    def get_proxy_class(cls):
        """
        Returns the proxy class. 

        :param config: Config object.
        :type config: bulbs.config.Config

        :rtype: class

        """
        return NodeProxy


    def save(self):
        """
        Saves/updates the element's data in the database.

        :rtype: None

        """
        data = self._get_property_data()
        index_name = self.get_index_name(self._client.config)
        keys = self.get_index_keys()
        self._client.update_indexed_vertex(self._id, data, index_name, keys)
        
    #
    # Override the _create and _update methods to cusomize behavior.
    #
    
    def _create(self, _data, kwds):  
        """
        Creates a vertex in the database; called by the NodeProxy create() method.

        :param _data: Optional property data dict.
        :type _data: dict

        :param kwds: Optional property data keyword pairs. 
        :type kwds: dict

        :rtype: None
        
        """
        # bundle is an OrderedDict containing data, index_name, and keys
        data, index_name, keys = self.get_bundle(_data, **kwds)
        resp = self._client.create_indexed_vertex(data, index_name, keys)
        result = resp.one()
        self._initialize(result)
        
    def _update(self, _id, _data, kwds):
        """
        Updates a vertex in the database; called by NodeProxy update() method.
        
        :param _id: Element ID.
        :param _id: int or str

        :param _data: Optional property data dict.
        :type _data: dict

        :param kwds: Optional property data keyword pairs. 
        :type kwds: dict

        :rtype: None
        
        """
        data, index_name, keys = self.get_bundle(_data, **kwds)
        resp = self._client.update_indexed_vertex(_id, data, index_name, keys)
        result = resp.one()
        self._initialize(result)
        
    def _initialize(self, result):
        """
        Initializes the element. Initialize all non-DB attributes here.

        :param result: Result object.
        :type result: Result

        :rtype: None

        ..note:: Called by _create, _update, and utils.initialize_element. 

        """
        Vertex._initialize(self,result)
        self._initialized = False
        self._set_property_data()
        self._initialized = True


class Relationship(Model, Edge):
    """ 
    Abstract base class used for creating a Relationship Model.
 
    It's used to create classes that model domain objects, and it's not meant 
    to be used directly. To use it, create a subclass specific to the type of 
    data you are storing. 

    Example usage for an edge between a blog entry node and its creating user::

        # people.py

        from bulbs.model import Relationship
        from bulbs.properties import DateTime
        from bulbs.utils import current_timestamp

        class Knows(Relationship):

            label = "knows"

            created = DateTime(default=current_timestamp, nullable=False)


    Example usage::

          >>> from people import Person, Knows
          >>> from bulbs.neo4jserver import Graph
          >>> g = Graph()

          # Add proxy interfaces to the Graph object for each custom Model
          >>> g.add_proxy("people", Person)
          >>> g.add_proxy("knows", Knows)

          # Create two Person nodes, which are automatically saved in the DB
          >>> james = g.people.create(name="James")
          >>> julie = g.people.create(name="Julie")

          # Create a "knows" relationship between James and Julie:
          >>> knows = g.knows.create(james,julie)
          >>> knows.timestamp

          # Get the people James knows (the outgoing vertices labeled "knows")
          >>> friends = james.outV('knows')

    """
    label = None

    @classmethod
    def get_label(cls, config):
        """
        Returns the edge's label.

        :param config: Config object.
        :type config: bulbs.config.Config

        :rtype: str

        """
        label = getattr(cls, config.label_var)
        return label

    @classmethod
    def get_element_key(cls, config):
        """
        Returns the element key.

        :param config: Config object.
        :type config: bulbs.config.Config

        :rtype: str

        """
        return cls.get_label(config)

    @classmethod 
    def get_index_name(cls, config):
        """
        Returns the index name. 

        :param config: Config object.
        :type config: bulbs.config.Config

        :rtype: str

        """
        return cls.get_label(config)

    @classmethod 
    def get_proxy_class(cls):
        """
        Returns the proxy class. 

        :param config: Config object.
        :type config: bulbs.config.Config

        :rtype: class

        """
        return RelationshipProxy


    def save(self):
        """
        Saves/updates the element's data in the database.

        :rtype: None

        """
        data = self._get_property_data()
        index_name = self.get_index_name(self._client.config)
        keys = self.get_index_keys()
        self._client.update_indexed_edge(self._id, data, index_name, keys)

    #
    # Override the _create and _update methods to customize behavior.
    #

    def _create(self, outV, inV, _data, kwds):
        """
        Creates an edge in the DB; called by RelatinshipProxy create() method.

        :param _data: Optional property data dict.
        :type _data: dict

        :param kwds: Optional property data keyword pairs. 
        :type kwds: dict

        :rtype: None
        
        """
        label = self.get_label(self._client.config)
        outV, inV = coerce_vertices(outV, inV)
        data, index_name, keys = self.get_bundle(_data, **kwds)
        resp = self._client.create_indexed_edge(outV, label, inV, data, index_name, keys)
        result = resp.one()
        self._initialize(result)
        
    def _update(self, _id, _data, kwds):
        """
        Updates an edge in DB; called by RelationshipProxy update() method.
        
        :param _id: Element ID.
        :param _id: int or str

        :param _data: Optional property data dict.
        :type _data: dict

        :param kwds: Optional property data keyword pairs. 
        :type kwds: dict

        :rtype: None
        
        """
        data, index_name, keys = self.get_bundle(_data, **kwds)
        resp = self._client.update_indexed_edge(_id, data, index_name, keys)
        result = resp.one()
        self._initialize(result)

    def _initialize(self,result):
        """
        Initializes the element. Initialize all non-DB attributes here.

        :param result: Result object.
        :type result: Result

        :rtype: None

        ..note:: Called by _create, _update, and utils.initialize_element. 

        """
        Edge._initialize(self,result)
        self._initialized = False
        self._set_property_data()
        self._initialized = True


class NodeProxy(VertexProxy):

    def create(self, _data=None, **kwds):
        """
        Adds a vertex to the database and returns it.

        :param _data: Optional property data dict.
        :type _data: dict

        :param kwds: Optional property data keyword pairs. 
        :type kwds: dict

        :rtype: Node

        """
        node = self.element_class(self.client)
        node._create(_data, kwds)
        return node
        
    def update(self, _id, _data=None, **kwds):
        """
        Updates an element in the graph DB and returns it.

        :param _id: The vertex ID.
        :type _id: int or str

        :param _data: Opetional property data dict.
        :type _data: dict

        :param kwds: Optional property data keyword pairs. 
        :type kwds: dict

        :rtype: Node

        """ 
        node = self.element_class(self.client)
        node._update(_id, _data, kwds)
        return node

    def get_all(self):
        """
        Returns all the elements for the model type.
        
        :rtype: Node generator
 
        """

        config = self.client.config
        type_var = config.type_var
        element_type = self.element_class.get_element_type(config)
        return self.index.lookup(type_var,element_type)

    def get_property_keys(self):
        """
        Returns a list of all the Property keys.

        :rtype: list

        """
        return self.element_class._properties.keys()

class RelationshipProxy(EdgeProxy):

    def create(self, outV, inV, _data=None, **kwds):
        """
        Creates an edge in the database and returns it.
        
        :param outV: The outgoing vertex. 
        :type outV: Vertex or int
              
        :param inV: The incoming vertex. 
        :type inV: Vertex or int

        :param _data: Optional property data dict.
        :type _data: dict

        :param kwds: Optional property data keyword pairs. 
        :type kwds: dict

        :rtype: Relationship

        """ 
        relationship = self.element_class(self.client)
        relationship._create(outV, inV, _data, kwds)
        return relationship

    def update(self, _id, _data=None, **kwds):
        """ 
        Updates an edge in the database and returns it. 
        
        :param _id: The edge ID.
        :type _id: int or str

        :param _data: Optional property data dict.
        :type _data: dict

        :param kwds: Optional property data keyword pairs. 
        :type kwds: dict

        :rtype: Relationship

        """
        relationship = self.element_class(self.client)
        relationship._update(_id, _data, kwds)
        return relationship

    def get_all(self):
        """
        Returns all the relationships for the label.

        :rtype: Relationship generator
 
        """
        # TODO: find a blueprints method that returns all edges for a given 
        # label because you many not want to index edges
        config = self.client.config
        label_var = config.label_var
        label = self.element_class.get_label(config)
        return self.index.lookup(label_var,label)


    def get_property_keys(self):
        """
        Returns a list of all the Property keys.

        :rtype: list

        """
        return self.element_class._properties.keys()
        



# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Bulbs supports plugabble type systems.

"""

class TypeSystem(object):
    """
    Abstract base class for plugabble database type systems.

    :cvar content_type: The backend client's content type.
    :cvar database: Converter object. Converts Python values to database values.
    :cvar python: Converter object. Converts database values to Python values.

    """
    content_type = None
    database = None
    python = None


class Converter(object):
    """Abstract base class of conversion methods called by DataType classes."""

    def to_string(self, value):
        raise NotImplementedError

    def to_integer(self, value):
        raise NotImplementedError
    
    def to_long(self, value):
        raise NotImplementedError

    def to_float(self, value):
        raise NotImplementedError

    def to_bool(self, value):
        raise NotImplementedError

    def to_list(self, value):
        raise NotImplementedError

    def to_dictionary(self, value):
        raise NotImplementedError

    def to_null(self, value):
        raise NotImplementedError

    def to_document(self, value):
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
import os
from .utils import get_logger, urlparse
from logging import StreamHandler, DEBUG, INFO, WARNING, ERROR, CRITICAL

log = get_logger(__name__)


class Config(object):
    """
    Configuration options for Bulbs.

    :param root_uri: Root URI of the database.
    :type root_uri: str

    :param username: Optional username. Defaults to None.
    :type username: str

    :param password: Optional password. Defaults to None.
    :type password: str

    :ivar root_uri: Root URI of the server.
    :ivar username: Optional username. Defaults to None.
    :ivar password: Optional password. Defaults to None.
    :ivar log_level: Python log level. Defaults to ERROR.
    :ivar log_handler: Python log handler. Defaults to StreamHandler.
    :ivar id_var: Name of the element ID variable. Defaults to "eid".
    :ivar type_var: Name of the type variable. Defaults to "element_type".
    :ivar label_var: Name of the label variable. Defaults to "label".
    :ivar type_system: Name of the type system. Defaults to "json".
    :ivar vertex_index: Name of the vertex index. Defaults to "vertex". 
    :ivar edge_index: Name of the edge index. Defaults to "edge". 
    :ivar autoindex: Enable auto indexing. Defaults to True.

    Example:

    >>> from bulbs.config import Config, DEBUG
    >>> from bulbs.neo4jserver import Graph, NEO4J_URI
    >>> config = Config(NEO4J_URI, username="james", password="secret")
    >>> config.set_logger(DEBUG)
    >>> g = Graph(config)

    """

    def __init__(self, root_uri, username=None, password=None):
        self.root_uri = root_uri
        self.username = username
        self.password = password
        self.log_level = ERROR
        self.log_handler = StreamHandler
        self.id_var = "eid"
        self.type_var = "element_type"
        self.label_var = "label"
        self.type_system = "json" 
        self.vertex_index = "vertex"
        self.edge_index = "edge"
        self.autoindex = True
        
        # Set the default log level and log handler
        self.set_logger(self.log_level, self.log_handler)

        # Sanity checks...
        assert self.root_uri is not None

    # TODO: fix duplicate log issue from setting logger multiple times
    def set_logger(self, log_level, log_handler=None):
        """
        Sets or updates the log level and log handler.

        :param log_level: Python log level.
        :type log_level: int

        :param log_handler: Python log handler. Defaults to log_handler.
        :type log_handler: logging.Handler

        :rtype: None

        """
        log = get_logger(__name__)
        log.setLevel(log_level)
        self.log_level = log_level 
        if log_handler is not None:
            log.root.addHandler(log_handler())

    def set_neo4j_heroku(self, log_level=ERROR, log_handler=None):
        """
        Sets credentials if using the Neo4j Heroku Add On.

        :param log_level: Python log level. Defaults to ERROR.
        :type log_level: int

        :param log_handler: Python log handler. Defaults to log_handler.
        :type log_handler: logging.Handler

        :rtype: None

        """
        url = os.environ.get('NEO4J_REST_URL', None)
        log.debug("NEORJ_REST_URL: %s", url)

        if url is not None:
            parsed =  urlparse(url)
            pieces = (parsed.scheme, parsed.hostname, parsed.port, parsed.path)
            self.root_uri = "%s://%s:%s%s" % pieces
            self.username = parsed.username
            self.password = parsed.password
            self.set_logger(log_level, log_handler)
            log.debug("ROOT_URI: %s", self.root_uri)
            log.debug("USERNAME: %s", self.username)
            log.debug("PASSWORD: %s", self.password)



import unittest
from bulbs.config import Config
#from bulbs.rest import Request

from bulbs.utils import build_path
from bulbs.rexster.client import RexsterRequest

class RestTestCase(unittest.TestCase):
    
    def setUp(self):
        #self.client = Client(config.DATABASE_URL)
        config = Config(root_uri=None)
        self.request = RexsterRequest(config)

    def test_init(self):
        config = Config('http://localhost:8182/not-graphs/gratefulgraph')
        assert config.root_uri == 'http://localhost:8182/not-graphs/gratefulgraph'

    def test_post(self):
        name_in = "james"
        email_in = "james@jamesthornton.com"

        path = build_path("vertices")
        params = dict(name=name_in, email=email_in)
        resp = self.request.post(path,params)
                
        assert resp.results._type == 'vertex'
        assert name_in == resp.results.get('name') 
        assert email_in == resp.results.get('email')

        # use the results of this function for get and delete tests 
        return resp
        
    def test_get(self):
        resp1 = self.test_post()
        oid1 = resp1.results.get('_id')
        element_type1 = resp1.results.get('_type')
        name1 = resp1.results.get('name')
        email1 = resp1.results.get('email')

        path = build_path("vertices",oid1)
        params = dict(name=name1, email=email1)
        resp2 = self.request.get(path,params)

        assert oid1 == resp2.results.get('_id')
        assert element_type1 == resp2.results.get('_type')
        assert name1 == resp2.results.get('name')
        assert email1 == resp2.results.get('email')

    def test_delete(self):

        resp1 = self.test_post()
        oid1 = resp1.results.get('_id')
        name1 = resp1.results.get('name')
        email1 = resp1.results.get('email')

        # make sure it's there
        #assert type(oid1) == int

        # delete it
        path = build_path("vertices",oid1)
        params = dict(name=name1, email=email1)
        resp2 = self.request.delete(path,params)

        # verify it's gone
        #assert "SNAPSHOT" in resp2.rexster_version
        #assert type(resp2.get('query_time')) == float
        assert resp2.results == None


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(RestTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='suite')




import unittest
from bulbs.config import Config
from bulbs.tests import BulbsTestCase, bulbs_test_suite
from bulbs.neo4jserver import Graph, Neo4jClient, NEO4J_URI, \
   VertexIndexProxy, EdgeIndexProxy, ExactIndex
from bulbs.tests import GremlinTestCase

config = Config(NEO4J_URI)
BulbsTestCase.client = Neo4jClient(config)
BulbsTestCase.vertex_index_proxy = VertexIndexProxy
BulbsTestCase.edge_index_proxy = EdgeIndexProxy
BulbsTestCase.index_class = ExactIndex
BulbsTestCase.graph = Graph(config)

def test_suite():
    suite = bulbs_test_suite()
    #suite.addTest(unittest.makeSuite(RestTestCase))
    suite.addTest(unittest.makeSuite(GremlinTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')




import os
import re
import sre_parse
import sre_compile
from sre_constants import BRANCH, SUBPATTERN
import hashlib
from . import utils

# GroovyScripts is the only public class

#
# The scanner code came from the TED project.
#

# TODO: Simplify this. You don't need group pattern detection.


class GroovyScripts(object):
    """
    Store and manage an index of Gremlin-Groovy scripts.

    :param file_path: Path to the base Groovy scripts file.
    :type file_path: str

    :ivar source_files: List containing the absolute paths to the script files,
                        in the order they were added.
    :ivar methods: Dict mapping Groovy method names to the actual scripts.

    .. note:: Use the update() method to add subsequent script files. 
              Order matters. Groovy methods are overridden if subsequently added
              files contain the same method name as a previously added file.

    """
    #: Relative path to the default script file
    default_file = "gremlin.groovy"

    def __init__(self, file_path=None):
        self.source_files = list()  # an ordered set might be better

        # methods format: methods[method_name] = method_body
        self.methods = dict()

        if file_path is None:
            file_path = self._get_default_file()
        self.update(file_path)

    def get(self, method_name):
        """
        Returns the Groovy script with the method name.
        
        :param method_name: Method name of a Groovy script.
        :type method_name: str

        :rtype: str

        """
        return self.methods[method_name]
        #script = self._build_script(method_definition, method_signature)
        #return script

    def update(self, file_path):
        """
        Updates the script index with the Groovy methods in the script file.

        :rtype: None

        """
        file_path = os.path.abspath(file_path)
        methods = self._get_methods(file_path)
        self._add_source_file(file_path)
        self.methods.update(methods)

    def refresh(self):
        """
        Refreshes the script index by re-reading the Groovy source files.

        :rtype: None

        """
        for file_path in self.source_files:
            methods = self._get_methods(file_path)
            self.methods.update(methods)

    def _add_source_file(self,file_path):
        # order matters (last in takes precedence if it overrides a method)
        self.source_files.append(file_path)

    def _get_methods(self,file_path):
        return Parser(file_path).get_methods()

    def _get_default_file(self):
        file_path = utils.get_file_path(__file__, self.default_file)
        return file_path

    def _build_script(definition, signature): 
        # This method isn't be used right now...
        # This method is not current (rework it to suit needs).
        script = """
        try {
          current_sha1 = methods[name]
        } catch(e) {
          current_sha1 = null
          methods = [:]
          methods[name] = sha1
        }
        if (current_sha1 == sha1) 
          %s

        try { 
          return %s
        } catch(e) {

          return %s 
        }""" % (signature, definition, signature)
        return script



class Scanner:
    def __init__(self, lexicon, flags=0):
        self.lexicon = lexicon
        self.group_pattern = self._get_group_pattern(flags)
        
    def _get_group_pattern(self,flags):
        # combine phrases into a compound pattern
        patterns = []
        sub_pattern = sre_parse.Pattern()
        sub_pattern.flags = flags
        for phrase, action in self.lexicon:
            patterns.append(sre_parse.SubPattern(sub_pattern, [
                (SUBPATTERN, (len(patterns) + 1, sre_parse.parse(phrase, flags))),
                ]))
        sub_pattern.groups = len(patterns) + 1
        group_pattern = sre_parse.SubPattern(sub_pattern, [(BRANCH, (None, patterns))])
        return sre_compile.compile(group_pattern)

    def get_multiline(self,f,m):
        content = []
        next_line = ''
        while not re.search("^}",next_line):
            content.append(next_line)
            try:
                next_line = next(f)    
            except StopIteration:
                # This will happen at end of file
                next_line = None
                break
        content = "".join(content)       
        return content, next_line

    def get_item(self,f,line):
        # IMPORTANT: Each item needs to be added sequentially 
        # to make sure the record data is grouped properly
        # so make sure you add content by calling callback()
        # before doing any recursive calls
        match = self.group_pattern.scanner(line).match() 
        if not match:
            return
        callback = self.lexicon[match.lastindex-1][1]
        if "def" in match.group():
            # this is a multi-line get
            first_line = match.group()
            body, current_line = self.get_multiline(f,match)
            sections = [first_line, body, current_line]
            content = "\n".join(sections).strip()
            callback(self,content)
            if current_line:
                self.get_item(f,current_line)
        else:
            callback(self,match.group(1))

    def scan(self,file_path):
        fin = open(file_path, 'r')    
        for line in fin:
            self.get_item(fin,line)

    
class Parser(object):

    def __init__(self, groovy_file):
        self.methods = {}
        # handler format: (pattern, callback)
        handlers = [ ("^def( .*)", self.add_method), ]
        Scanner(handlers).scan(groovy_file)

    def get_methods(self):
        return self.methods

    # Scanner Callback
    def add_method(self,scanner,token):
        method_definition = token
        method_signature = self._get_method_signature(method_definition)
        method_name = self._get_method_name(method_signature)
        method_body = self._get_method_body(method_definition)
        # NOTE: Not using sha1, signature, or the full method right now
        # because of the way the GSE works. It's easier to handle version
        # control by just using the method_body, which the GSE compiles,
        # creates a class out of, and stores in a classMap for reuse.
        # You can't do imports inside Groovy methods so just using the func body 
        #sha1 = self._get_sha1(method_definition)
        #self.methods[method_name] = (method_signature, method_definition, sha1)
        self.methods[method_name] = method_body

    def _get_method_signature(self,method_definition):
        pattern = '^def(.*){'
        return re.search(pattern,method_definition).group(1).strip()
            
    def _get_method_name(self,method_signature):
        pattern = '^(.*)\('
        return re.search(pattern,method_signature).group(1).strip()

    def _get_method_body(self,method_definition):
        # remove the first and last lines, and return just the method body
        lines = method_definition.split('\n')
        body_lines = lines[+1:-1]
        method_body = "\n".join(body_lines).strip()
        return method_body

    def _get_sha1(self,method_definition):
        # this is used to detect version changes
        sha1 = hashlib.sha1()
        sha1.update(method_definition)
        return sha1.hexdigest()




#print Parser("gremlin.groovy").get_methods()



# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
The JSON Type System.

"""
# Python 3
import six
import sys
if sys.version > '3':
    long = int
    unicode = str

from bulbs.base import TypeSystem, Converter
from .utils import to_timestamp, to_datetime, to_datestamp, to_date, json


class DatabaseConverter(Converter):
    """
    Converts Python values to database values.

    .. note:: Python to JSON conversion is usually just a simple pass through.

    """
    def to_string(self, value):
        """
        Converts a Python byte string to a unicode string.
        
        :param value: Property value. 
        :type value: str or None

        :rtype: unicode or None

        :raises: ValueError

        """
        # NOTE: Using unicode instead of str
        if value is not None:
            return unicode(value)

    def to_integer(self, value):
        """
        Passes through a Python integer.

        :param value: Property value. 
        :type value: int or None

        :rtype: int or None

        """
        return value
    
    def to_long(self, value):
        """
        Passes through a Python long.

        :param value: Property value. 
        :type value: long or None

        :rtype: long or None

        """
        return value

    def to_bool(self, value):
        """
        Passes through a Python bool.

        :param value: Property value.
        :type value: bool or None

        :rtype: bool or None

        """
        return value

    def to_float(self, value):
        """
        Passes through a Python float.

        :param value: Property value. 
        :type value: float or None

        :rtype: float or None

        """
        return value

    def to_list(self, value):
        """
        Passes through a Python list.

        :param value: Property value. 
        :type value: list or None

        :rtype: list or None

        """
        return value

    def to_dictionary(self, value):
        """
        Passes through a Python dictionary.

        :param value: Property value. 
        :type value: dict or None

        :rtype: dict or None

        """
        return value

    def to_datetime(self, value):
        """
        Converts a Python datetime object to a timestamp integer.

        :param value: Property value. 
        :type value: datetime or None

        :rtype: int or None

        """
        if value is not None:
            return to_timestamp(value)

    def to_date(self, value):
        """
        Converts a Python date object to a timestamp integer.

        :param value: Property value. 
        :type value: date or None

        :rtype: int or None

        """
        if value is not None:
            return to_datestamp(value)

    def to_null(self, value):
        """
        Passes through a Python None.

        :param value: Property value. 
        :type value: None

        :rtype: None

        """
        return value

    def to_document(self, value):
        """
        Converts a Python object to a json string

        :param value: Property value.
        :type value: dict or list or None

        :rtype: unicode or None
        """
        if value is not None:
            return unicode(json.dumps(value))


class PythonConverter(Converter):
    """Converts database values to Python values."""

    # TODO: Why are we checking if value is not None?
    # This is supposed to be handled elsewhere.
    # Conversion exceptions are now handled in Property.convert_to_python() 
    
    def to_string(self, value):
        """
        Converts a JSON string to a Python unicode string.
        
        :param value: Property value. 
        :type value: str or None

        :rtype: unicode or None

        :raises: ValueError

        """
        if value is not None:
            return unicode(value)

    def to_integer(self, value):
        """
        Converts a JSON number to a Python integer.

        :param value: Property value. 
        :type value: int or None

        :rtype: int or None

        :raises: ValueError

        """
        if value is not None:
            return int(value)

    def to_long(self, value):
        """
        Converts a JSON number to a Python long.

        :param value: Property value. 
        :type value: long or None

        :rtype: long or None

        :raises: ValueError

        """
        if value is not None:
            return long(value)

    def to_float(self, value):
        """
        Converts a JSON number to a Python float.

        :param value: Property value. 
        :type value: float or None

        :rtype: float or None

        :raises: ValueError

        """
        if value is not None:
            return float(value)              

    def to_bool(self, value):
        """
        Converts a JSON boolean value to a Python bool.

        :param value: Property value.
        :type value: bool or None

        :rtype: bool or None

        :raises: ValueError

        """
        if value is not None:
            return bool(value)

    def to_list(self, value):
        """
        Converts a JSON list to a Python list.

        :param value: Property value. 
        :type value: list or None

        :rtype: list or None

        :raises: ValueError

        """
        if value is not None:
            return list(value)

    def to_dictionary(self, value):
        """
        Converts a JSON map to a Python dictionary.         

        :param value: Property value. 
        :type value: dict or unicode or None

        :rtype: dict or None

        :raises: ValueError

        """
        if value is None:
            return None

        if isinstance(value, unicode):
            return json.loads(value)

        return dict(value)

    def to_datetime(self, value):
        """
        Converts a JSON integer timestamp to a Python datetime object.

        :param value: Property value. 
        :type value: int or None

        :rtype: datetime or None

        :raises: ValueError

        """
        if value is not None:
            return to_datetime(value)
            
    def to_date(self, value):
        """
        Converts a JSON integer timestamp to a Python date object.

        :param value: Property value. 
        :type value: int or None

        :rtype: date or None

        :raises: ValueError

        """
        if value is not None:
            return to_date(value)

    def to_null(self, value):
        """
        Converts a JSON null to a Python None.

        :param value: Property value. 
        :type value: None

        :rtype: None

        :raises: ValueError

        """
        if value is not None:
            raise ValueError

        return None


class JSONTypeSystem(TypeSystem):
    """
    Converts database properties to and from their JSON representations.

    :cvar content_type: The backend client's content type.
    :cvar database: Converter object. Converts Python values to database values.
    :cvar python: Converter object. Converts database values to Python values.

    """
    content_type = "application/json"

    database = DatabaseConverter()
    python = PythonConverter()
    














import unittest
from .client_tests import neo4j_client_suite
from .bulbs_tests import test_suite as bulbs_test_suite


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(neo4j_client_suite())
    suite.addTest(bulbs_test_suite())
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='suite')




# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Bulbs supports pluggable clients. This is the Rexster client.

"""
from bulbs.config import Config, DEBUG
from bulbs.registry import Registry
from bulbs.utils import get_logger

# specific to this client
from bulbs.json import JSONTypeSystem
from bulbs.base import Client, Response, Result 
from bulbs.rest import Request, RESPONSE_HANDLERS
from bulbs.groovy import GroovyScripts

from bulbs.utils import json, build_path, get_file_path, urlsplit, coerce_id


##### Titan

from bulbs.rexster.client import RexsterClient, \
    RexsterResponse, RexsterResult

# The default URIs
TITAN_URI = "http://localhost:8182/graphs/titanexample"
SAIL_URI = "http://localhost:8182/graphs/sailgraph"

# The logger defined in Config
log = get_logger(__name__)

# Rexster resource paths
# TODO: local path vars would be faster
vertex_path = "vertices"
edge_path = "edges"
index_path = "indices"
gremlin_path = "tp/gremlin"
transaction_path = "tp/batch/tx"
multi_get_path = "tp/batch"
key_index_path = "keyindices"

class TitanResult(RexsterResult):
    """
    Container class for a single result, not a list of results.

    :param result: The raw result.
    :type result: dict

    :param config: The client Config object.
    :type config: Config 

    :ivar raw: The raw result.
    :ivar data: The data in the result.

    """
    pass


class TitanResponse(RexsterResponse):
    """
    Container class for the server response.

    :param response: httplib2 response: (headers, content).
    :type response: tuple

    :param config: Config object.
    :type config: bulbs.config.Config

    :ivar config: Config object.
    :ivar headers: httplib2 response headers, see:
        http://httplib2.googlecode.com/hg/doc/html/libhttplib2.html
    :ivar content: A dict containing the HTTP response content.
    :ivar results: A generator of RexsterResult objects, a single RexsterResult object, 
        or None, depending on the number of results returned.
    :ivar total_size: The number of results returned.
    :ivar raw: Raw HTTP response. Only set when log_level is DEBUG.

    """
    result_class = TitanResult



class TitanRequest(Request):
    """Makes HTTP requests to Rexster and returns a RexsterResponse.""" 
    
    response_class = TitanResponse


class TitanClient(RexsterClient):
    """
    Low-level client that sends a request to Titan and returns a response.

    :param config: Optional Config object. Defaults to default Config.
    :type config: bulbs.config.Config

    :cvar default_uri: Default URI for the database.
    :cvar request_class: Request class for the Client.

    :ivar config: Config object.
    :ivar registry: Registry object.
    :ivar scripts: GroovyScripts object.  
    :ivar type_system: JSONTypeSystem object.
    :ivar request: TitanRequest object.

    Example:

    >>> from bulbs.titan import TitanClient
    >>> client = TitanClient()
    >>> script = client.scripts.get("get_vertices")
    >>> response = client.gremlin(script, params=None)
    >>> result = response.results.next()

    """ 
    #: Default URI for the database.
    default_uri = TITAN_URI
    request_class = TitanRequest

    def __init__(self, config=None, db_name=None):
        super(TitanClient, self).__init__(config, db_name)

    # Titan-Specific Index Methods

    def create_vertex_key_index(self, key):
        path = build_path(key_index_path, "vertex", key)
        params = None
        return self.request.post(path, params)

    def create_edge_key_index(self, key):
        # Titan doesn't support edge indices
        #path = build_path(key_index_path, "edge", key)
        #params = None
        #return self.request.post(path, params)
        raise NotImplementedError

    # GET 

    # these could replace the Rexster Gremlin version of these methods
    def outV(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "out")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)
    
    def inV(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "in")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    def bothV(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "both")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    def outV_count(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "outCount")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    def inV_count(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "inCount")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    def bothV_count(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "bothCount")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    def outV_ids(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "outIds")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    def inV_ids(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "inIds")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    def bothV_ids(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "bothIds")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    def outE(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "outE")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)
    
    def inE(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "inE")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    def bothE(self, _id, label=None, start=None, limit=None, properties=None):
        path = build_path(vertex_path, _id, "bothE")
        params = build_params(_label=label, _limit=limit, _properties=properties)
        return self.request.get(path, params)

    # Key Indices

    def get_vertex_keys(self):
        path = build_path(key_index_path, "vertex")
        params = None
        return self.request.get(path, params)

    def get_edge_keys(self):
        path = build_path(key_index_path, "edge")
        params = None
        return self.request.get(path, params)

    def get_all_keys(self):
        path = key_index_path
        params = None
        return self.request.get(path, params)

    # POST
    
    def create_vertex_index_key(self, key):
        path = build_path(key_index_path, "vertex", key)
        params = None
        return self.request.post(path, params)

    def create_edge_index_key(self, key):
        path = build_path(key_index_path, "edge", key)
        params = None
        return self.request.post(path, params)


    # Index Proxy - General

    def get_all_indices(self):
        """Returns a list of all the element indices."""
        raise NotImplementedError

    def get_index(self, name):
        raise NotImplementedError

    def delete_index(self, name): 
        raise NotImplementedError
    
    # Index Proxy - Vertex

    def create_vertex_index(self, index_name, *args, **kwds):
        """
        Creates a vertex index with the specified params.

        :param index_name: Name of the index to create.
        :type index_name: str

        :rtype: TitanResponse

        """
        raise NotImplementedError

    def get_vertex_index(self, index_name):
        """
        Returns the vertex index with the index_name.

        :param index_name: Name of the index.
        :type index_name: str

        :rtype: TitanResponse

        """
        raise NotImplementedError

    def get_or_create_vertex_index(self, index_name, index_params=None):
        raise NotImplementedError
        
    def delete_vertex_index(self, name): 
        """
        Deletes the vertex index with the index_name.

        :param index_name: Name of the index.
        :type index_name: str

        :rtype: TitanResponse

        """
        raise NotImplementedError

    # Index Proxy - Edge
    # Titan does NOT support edge indices

    def create_edge_index(self, name, *args, **kwds):
        raise NotImplementedError
        
    def get_edge_index(self, name):
        """
        Returns the edge index with the index_name.

        :param index_name: Name of the index.
        :type index_name: str

        :rtype: TitanResponse

        """
        raise NotImplementedError
        
    def get_or_create_edge_index(self, index_name, index_params=None):
        raise NotImplementedError

    def delete_edge_index(self, name):
        raise NotImplementedError

    # Index Container - Vertex

    def put_vertex(self, index_name, key, value, _id):
        # Titan only supports automatic indices
        raise NotImplementedError

    def lookup_vertex(self, index_name, key, value):
        """
        Returns the vertices indexed with the key and value.

        :param index_name: Name of the index.
        :type index_name: str

        :param key: Name of the key.
        :type key: str

        :param value: Value of the key.
        :type value: str

        :rtype: TitanResponse

        """
        # NOTE: this is different than Rexster's version
        # it uses vertex_path instead of index_path, and 
        # index_name is N/A
        # Keeping method interface the same for practical reasons so
        # index_name will be ignored, any value will work.
        path = build_path(vertex_path)
        params = dict(key=key,value=value)
        return self.request.get(path,params)

    def query_vertex(self, index_name, params):
        """Queries for an vertex in the index and returns the Response."""
        path = build_path(index_path,index_name)
        return self.request.get(path,params)

    def remove_vertex(self,index_name,_id,key=None,value=None):
        # Titan only supports automatic indices
        raise NotImplementedError

    # Index Container - Edge 
    # Titan does NOT support edge indices

    def put_edge(self, index_name, key, value, _id):
        raise NotImplementedError

    def lookup_edge(self, index_name, key, value):
        """
        Looks up an edge in the index and returns the Response.
        """
        # NOTE: this is different than Rexster's version
        # it uses edge_path instead of index_path, and 
        # index_name is N/A
        # Keeping method interface the same for practical reasons so
        # index_name will be ignored, any value will work.
        #path = build_path(edge_path)
        #params = dict(key=key,value=value)
        #return self.request.get(path,params)
        raise NotImplementedError

    def query_edge(self, index_name, params):
        """Queries for an edge in the index and returns the Response."""
        raise NotImplementedError

    def remove_edge(self, index_name, _id, key=None, value=None):
        raise NotImplementedError
    
    # Model Proxy - Vertex

    def create_indexed_vertex(self, data, index_name, keys=None):
        """
        Creates a vertex, indexes it, and returns the Response.

        :param data: Property data.
        :type data: dict

        :param index_name: Name of the index.
        :type index_name: str

        :param keys: Property keys to index.
        :type keys: list

        :rtype: TitanResponse

        """
        return self.create_vertex(data)
    
    def update_indexed_vertex(self, _id, data, index_name, keys=None):
        """
        Updates an indexed vertex and returns the Response.

        :param index_name: Name of the index.
        :type index_name: str

        :param data: Property data.
        :type data: dict

        :param index_name: Name of the index.
        :type index_name: str

        :param keys: Property keys to index.
        :type keys: list

        :rtype: TitanResponse

        """
        return self.update_vertex(_id, data)

    # Model Proxy - Edge

    def create_indexed_edge(self, outV, label, inV, data, index_name, keys=None):
        """
        Creates a edge, indexes it, and returns the Response.

        :param outV: Outgoing vertex ID.
        :type outV: int

        :param label: Edge label.
        :type label: str

        :param inV: Incoming vertex ID.
        :type inV: int

        :param data: Property data.
        :type data: dict

        :param index_name: Name of the index.
        :type index_name: str

        :param keys: Property keys to index. Defaults to None (indexes all properties).
        :type keys: list

        :rtype: TitanResponse

        """
        return self.create_edge(outV, label, inV, data)
        
    def update_indexed_edge(self, _id, data, index_name, keys=None):
        """
        Updates an indexed edge and returns the Response.

        :param _id: Edge ID.
        :type _id: int

        :param data: Property data.
        :type data: dict

        :param index_name: Name of the index.
        :type index_name: str

        :param keys: Property keys to index. Defaults to None (indexes all properties).
        :type keys: list

        :rtype: TitanResponse

        """
        return self.update_edge(_id, data)



# Utils

def build_params(**kwds):
    # Rexster isn't liking None param values
    params = dict()
    for key in kwds:
        value = kwds[key]
        if value is not None:
            params[key] = value
    return params



import sys
import unittest
import argparse
from bulbs.config import Config, DEBUG
from bulbs.tests import BulbsTestCase, bulbs_test_suite
from bulbs.titan import Graph, TitanClient, TITAN_URI, \
    VertexIndexProxy, EdgeIndexProxy, KeyIndex
from bulbs.tests import GremlinTestCase

# Setting a module var looks to be the easiest way to do this
db_name = "tinkergraph"

def test_suite():
    # pass in a db_name to test a specific database
    client = TitanClient(db_name=db_name)
    BulbsTestCase.client = client
    BulbsTestCase.vertex_index_proxy = VertexIndexProxy
    BulbsTestCase.edge_index_proxy = EdgeIndexProxy
    BulbsTestCase.index_class = KeyIndex
    BulbsTestCase.graph = Graph(client.config)

    suite = bulbs_test_suite()
    #suite.addTest(unittest.makeSuite(RestTestCase))
    suite.addTest(unittest.makeSuite(GremlinTestCase))
    return suite

if __name__ == '__main__':

    # TODO: Bubble up the command line option to python setup.py test.
    # http://eli.thegreenplace.net/2011/08/02/python-unit-testing-parametrized-test-cases/
    # http://www.doughellmann.com/PyMOTW/argparse/

    parser = argparse.ArgumentParser()
    parser.add_argument('--db', default="titanexample")
    args = parser.parse_args()
    
    db_name = args.db

    # Now set the sys.argv to the unittest_args (leaving sys.argv[0] alone)
    sys.argv[1:] = []

    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
An interface for interacting with indices on Rexster.

"""
from bulbs.utils import initialize_element, initialize_elements, get_one_result


class IndexProxy(object):
    """Abstract base class the index proxies."""

    def __init__(self, index_class, client):        
        # The index class for this proxy, e.g. ManualIndex.
        self.index_class = index_class

        # The Client object for the database.
        self.client = client
    

class VertexIndexProxy(IndexProxy):
    """
    Manage vertex indices on Rexster.

    :param index_class: The index class for this proxy, e.g. ManualIndex.
    :type index_class: Index

    :param client: The Client object for the database.
    :type client: bulbs.rexster.client.RexsterClient

    :ivar index_class: Index class.
    :ivar client: RexsterClient object.

    """
                        
    def create(self, index_name):
        """
        Creates an Vertex index and returns it.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index
        
        """
        raise NotImplementedError

    def get(self, index_name="vertex"):
        """
        Returns the Index object with the specified name or None if not found.
        
        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index
        
        """
        index = self.index_class(self.client, None)
        index.base_type = "vertex"
        index._index_name = index_name
        self.client.registry.add_index(index_name, index)
        return index

    def get_or_create(self, index_name="vertex", index_params=None):
        """
        Get a Vertex Index or create it if it doesn't exist.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index

        """ 
        return self.get(index_name)

    def delete(self, index_name):
        """ 
        Deletes an index and returns the Response.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.client.RexsterResponse

        """
        raise NotImplementedError


class EdgeIndexProxy(IndexProxy):
    """
    Manage edge indices on Rexster.

    :param index_class: The index class for this proxy, e.g. ManualIndex.
    :type index_class: Index

    :param client: The Client object for the database.
    :type client: bulbs.rexster.client.RexsterClient

    :ivar index_class: Index class.
    :ivar client: RexsterClient object.

    """

    def create(self,index_name,*args,**kwds):
        """ 
        Adds an index to the database and returns it. 

        index_keys must be a string in this format: '[k1,k2]'
        Don't pass actual list b/c keys get double quoted.

        :param index_name: The name of the index to create.

        :param index_class: The class of the elements stored in the index. 
                            Either vertex or edge.
        
        """
        raise NotImplementedError

    def get(self, index_name="edge"):
        """
        Returns the Index object with the specified name or None if not found.
        
        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index
        
        """
        index = self.index_class(self.client, None)
        index.base_type = "edge"
        index._index_name = index_name
        self.client.registry.add_index(index_name, index)
        return index


    def get_or_create(self, index_name="edge", index_params=None):
        """
        Get an Edge Index or create it if it doesn't exist.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index

        """ 
        return self.get(index_name)

    def delete(self,index_name):
        """ 
        Deletes an index and returns the Response.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.client.RexsterResponse

        """
        raise NotImplementedError

#
# Index Containers (Titan only supports KeyIndex so far)
#

class Index(object):
    """Abstract base class for an index."""

    def __init__(self, client, result):
        self.client = client
        self.result = result
        self.base_type = None # set by Factory.get_index
        self._index_name = None # ditto 
        # the index_name is actually ignored with Titan, 
        # but setting it like normal to make tests pass

    @classmethod 
    def get_proxy_class(cls, base_type):
        """
        Returns the IndexProxy class.

        :param base_type: Index base type, either vertex or edge.
        :type base_type: str

        :rtype: class

        """
        class_map = dict(vertex=VertexIndexProxy, edge=EdgeIndexProxy)
        return class_map[base_type]

    @property
    def index_name(self):
        """
        Returns the index name.

        :rtype: str

        """
        # faking the index name as "vertex"
        return self._index_name

    @property
    def index_class(self):
        """
        Returns the index class, either vertex or edge.

        :rtype: class

        """
        return self.base_type
    
    @property
    def index_type(self):
        """
        Returns the index type, which will either be automatic or manual.

        :rtype: str

        """
        return "automatic"

    def count(self,key=None,value=None,**pair):
        """
        Return a count of all elements with 'key' equal to 'value' in the index.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param pair: Optional keyword param. Instead of supplying key=name 
                     and value = 'James', you can supply a key/value pair in
                     the form of name='James'.
        """
        raise NotImplementedError


    def _get_key_value(self, key, value, pair):
        """Return the key and value, regardless of how it was entered."""
        if pair:
            key, value = pair.popitem()
        return key, value

    def _get_method(self, **method_map):
        method_name = method_map[self.index_class]
        method = getattr(self.client, method_name)
        return method

    def lookup(self, key=None, value=None, **pair):
        """
        Return a generator containing all the elements with key property equal 
        to value in the index.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize the results. Defaults to False. 

        :param pair: Optional keyword param. Instead of supplying key=name 
                     and value = 'James', you can supply a key/value pair in
                     the form of name='James'.
        """
        key, value = self._get_key_value(key, value, pair)
        resp = self.client.lookup_vertex(self.index_name,key,value)
        return initialize_elements(self.client,resp)


class KeyIndex(Index):

    def keys(self):
        """Return the index's keys."""
        # Titan does not support edge indices.
        resp = self.client.get_vertex_keys()
        return [result.raw for result in resp.results]

    def create_key(self, key):
        # TODO: You can't create a key if prop already exists - workaround?
        if self.base_type is "edge":
            # Titan does not support edge indices.
            raise NotImplementedError
        return self.client.create_vertex_index_key(key)

    def rebuild(self):
        raise NotImplementedError # (for now)
        # need class_map b/c the Blueprints need capitalized class names, 
        # but Rexster returns lower-case class names for index_class
        method_map = dict(vertex=self.client.rebuild_vertex_index,
                          edge=self.client.rebuild_edge_index)
        rebuild_method = method_map.get(self.index_class)
        resp = rebuild_method(self.index_name)
        return list(resp.results)
        

 



import unittest
import random

from bulbs.config import Config, DEBUG, ERROR
from bulbs.registry import Registry
from bulbs.base import TypeSystem

class ClientTestCase(unittest.TestCase):
    
    def setUp(self):
        self.client = None
        raise NotImplementedError

    def test_init(self):
        
        assert self.client.default_uri is not None
        assert isinstance(self.client.config, Config) 
        assert isinstance(self.client.registry, Registry)
        assert isinstance(self.client.type_system, TypeSystem)

    # Vertex Proxy

    def test_create_vertex(self):
        name, age = "James", 34
        data = dict(name=name, age=age)
        resp = self.client.create_vertex(data)
        assert resp.results.get_type() == "vertex"
        assert resp.results.get_data() == data  
        assert resp.results.data.get('name') == name
        assert resp.results.data.get('age') == age
  
    def test_get_vertex(self):
        name, age = "James", 34
        data = dict(name=name,age=age)
        resp1 = self.client.create_vertex(data)
        resp2 = self.client.get_vertex(resp1.results.get_id())
        assert resp1.results.data == resp2.results.data

    def test_get_all_vertices(self):
        resp = self.client.get_all_vertices()
        assert resp.total_size > 1

    def test_update_vertex(self):
        name1, age1 = "James", 34
        data1 = dict(name=name1, age=age1)
        resp1 = self.client.create_vertex(data1)        

        name2, age2 = "Julie", 28
        data2 = dict(name=name2,age=age2)
        resp2 = self.client.update_vertex(resp1.results.get_id(), data2)

        resp3 = self.client.get_vertex(resp1.results.get_id())

        assert resp3.results.get_type() == "vertex"
        assert resp3.results.data.get('name') == name2
        assert resp3.results.data.get('age') == age2

    def test_delete_vertex(self):
        name, age = "James", 34
        data = dict(name=name,age=age)
        resp1 = self.client.create_vertex(data)        

        deleted = False
        resp2 = self.client.delete_vertex(resp1.results.get_id())
        try:
            resp3 = self.client.get_vertex(resp1.results.get_id())
        except LookupError:
            deleted = True
        assert deleted is True

    # Edges
    def test_create_edge(self):
        resp1 = self.client.create_vertex({'name':'James','age':34})
        resp2 = self.client.create_vertex({'name':'Julie','age':28})

        outV = resp1.results.get_id()
        inV = resp2.results.get_id()
        label = "knows"
        data = dict(timestamp=123456789)
        resp3 = self.client.create_edge(outV, label, inV, data)

        assert resp3.results.get_type() == "edge"
        assert resp3.results.get_label() == label
        assert resp3.results.get_outV() == outV
        assert resp3.results.get_inV() == inV
        assert resp3.results.get_data() == data

    def test_get_edge(self):
        resp1 = self.client.create_vertex({'name':'James','age':34})
        resp2 = self.client.create_vertex({'name':'Julie','age':28})

        outV = resp1.results.get_id()
        inV = resp2.results.get_id()
        label = "knows"

        resp3 = self.client.create_edge(outV,label,inV)
        resp4 = self.client.get_edge(resp3.results.get_id())

        assert resp3.results.get_id() == resp4.results.get_id()
        assert resp3.results.get_type() == resp4.results.get_type()
        assert resp3.results.get_data() == resp4.results.get_data()

    def test_get_all_edges(self):
        resp1 = self.client.create_vertex({'name':'James','age':34})
        resp2 = self.client.create_vertex({'name':'Julie','age':28})

        outV = resp1.results.get_id()
        inV = resp2.results.get_id()
        label = "knows"
        data = dict(timestamp=123456789)
        resp3 = self.client.create_edge(outV, label, inV, data)
        resp4 = self.client.create_edge(inV, label, outV, data)

        resp = self.client.get_all_edges()
        assert resp.total_size > 1

    def test_update_edge(self):
        resp1 = self.client.create_vertex({'name':'James','age':34})
        resp2 = self.client.create_vertex({'name':'Julie','age':28})

        outV = resp1.results.get_id()
        inV = resp2.results.get_id()
        label = "knows"
        resp3 = self.client.create_edge(outV,label,inV)

        data = dict(timestamp=12345678)
        resp4 = self.client.update_edge(resp3.results.get_id(),data)

        assert resp4.results.get_data() == data

    def test_delete_edge(self):
        resp1 = self.client.create_vertex({'name':'James','age':34})
        resp2 = self.client.create_vertex({'name':'Julie','age':28})
        outV = resp1.results.get_id()
        inV = resp2.results.get_id()
        label = "knows"
        resp3 = self.client.create_edge(outV,label,inV)

        deleted = False
        resp4 = self.client.delete_edge(resp3.results.get_id())
        try:
            resp4 = self.client.get_edge(resp3.results.get_id())
        except LookupError:
            deleted = True
        assert deleted is True


    def test_vertex_container(self):
        resp1 = self.client.create_vertex({'name':'James','age':34})
        resp2 = self.client.create_vertex({'name':'Julie','age':28})

        outV = vertex_id1 =resp1.results.get_id()
        inV = vertex_id2 = resp2.results.get_id()
        label = "knows"

        resp3 = self.client.create_edge(outV, label, inV)
        edge_id = resp3.results.get_id()

        # Get the outgoing edge of vertex1
        outE = self.client.outE(vertex_id1).one()
        assert outE.get_id() == edge_id

        # Get the incoming edge of vertex2
        inE = self.client.inE(vertex_id2).one()
        assert inE.get_id() == edge_id

        # Get the incoming and outgoing edges of vertex1
        bothE = self.client.outE(vertex_id1).one()
        assert bothE.get_id() == edge_id
        
        # Get the outgoing edge of vertex1
        outV = self.client.outV(vertex_id1).one()
        assert outV.get_id() == vertex_id2

        # Get the incoming edge of vertex2
        inV = self.client.inV(vertex_id2).one()
        assert inV.get_id() == vertex_id1

        # Get the incoming and outgoing edges of vertex1
        bothV = self.client.outV(vertex_id1).one()
        assert bothV.get_id() == vertex_id2


#
# NOTE: client index tests moved to client_index_tests.py
#       because Titan does indexing differently - JT 10/22/2012
#    





import sys
import argparse

import unittest
from bulbs.config import Config
from bulbs.tests.client_tests import ClientTestCase
from bulbs.tests.client_index_tests import ClientIndexTestCase
from bulbs.titan.client import TitanClient, TITAN_URI, DEBUG

import time

# Default database. You can override this with the command line:
# $ python client_tests.py --db mydb
db_name = "titanexample"

#client = TitanClient(db_name=db_name)
#client.config.set_logger(DEBUG)

class TitanClientTestCase(ClientTestCase):

    def setUp(self):
        self.client = TitanClient(db_name=db_name)

# Separated client index tests for Titan
class TitanClientIndexTestCase(ClientIndexTestCase):

    def setUp(self):
        self.client = TitanClient(db_name=db_name)
        
    def test_create_vertex_index(self):
        pass

    def test_get_vertex_index(self):
        pass

    def test_indexed_vertex_CRUD(self):
        pass

    def test_indexed_edge_CRUD(self):
        pass


def titan_client_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TitanClientTestCase))
    suite.addTest(unittest.makeSuite(TitanClientIndexTestCase))
    return suite

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('--db', default=None)
    args = parser.parse_args()
    
    db_name = args.db

    # Now set the sys.argv to the unittest_args (leaving sys.argv[0] alone)
    sys.argv[1:] = []
    
    unittest.main(defaultTest='titan_client_suite')



import os
import yaml 
from string import Template
from .utils import get_file_path

# You only need this for Gremlin scripts if the server doesn't implement param 
# bindings; otherwise, use groovy.py with gremlin.groovy -- it's several 
# hunderd times faster. Cypher is handled differently on the server side so 
# there there is no performance issue.

class Yaml(object):
    """Load Gremlin or Cypher YAML templates from a .yaml source file."""

    def __init__(self,file_name=None):
        self.file_name = self._get_file_name(file_name)
        self.templates = self._get_templates(self.file_name)

    def get(self,name,params={}):
        """Return a Gremlin script or Cypher query, generated from the params."""
        template = self.templates.get(name)
        #params = self._quote_params(params)
        return template.substitute(params)

    def update(self,file_name):
        new_templates = self._get_templates(file_name)
        self.templates.update(new_templates)
        
    def refresh(self):
        """Refresh the stored templates from the YAML source."""
        self.templates = self._get_templates()

    def _get_file_name(self,file_name):
        if file_name is None:
            dir_name = os.path.dirname(__file__)
            file_name = get_file_path(dir_name,"gremlin.yaml")
        return file_name

    def _get_templates(self,file_name):
        templates = dict()
        f = open(file_name)
        yaml_map = yaml.load(f)    
        for name in yaml_map: # Python 3
            template = yaml_map[name] 
            templates[name] = Template(template)
        return templates

    #def _quote_params(self,params):
    #    quoted_tuple = map(self._quote,params.items())
    #    params = dict(quoted_tuple)
    #    return params

    #def _quote(self,pair):
    #    key, value = pair
    #    if type(value) == str:
    #        value = "'%s'" % value
    #    elif value is None:
    #        value = ""
    #    return key, value



import unittest
from bulbs.config import Config
from bulbs.utils import json
from bulbs.neo4jserver import Neo4jClient, NEO4J_URI
from bulbs.tests.client_tests import ClientTestCase
from bulbs.tests.client_index_tests import ClientIndexTestCase

from bulbs.factory import Factory
from bulbs.element import Vertex, Edge
from bulbs.neo4jserver.index import ExactIndex

import time


class Neo4jClientTestCase(ClientTestCase):

    def setUp(self):
        config = Config(NEO4J_URI)
        self.client = Neo4jClient(config)

# Separated client index tests for Titan
class Neo4jClientIndexTestCase(ClientIndexTestCase):

    def setUp(self):
        config = Config(NEO4J_URI)
        self.client = Neo4jClient(config)


# why is this here? - JT 10/22/2012
class Neo4jIndexTestCase(unittest.TestCase):
    
    def setUp(self):
        config = Config(NEO4J_URI)
        self.client = Neo4jClient(config)
        self.factory = Factory(self.client)

    def test_gremlin(self):
        # limiting return count so we don't exceed heap size
        resp = self.client.gremlin("g.V[0..9]")
        assert resp.total_size > 5

    def test_query_exact_vertex_index(self):
        index = self.factory.get_index(Vertex, ExactIndex)
        vertices = index.query("name", "Jam*")
        assert len(list(vertices)) > 1

    def test_query_exact_edge_index(self):
        index = self.factory.get_index(Edge, ExactIndex)
        edges = index.query("timestamp", "1*")
        assert len(list(edges)) > 1


class CypherTestCase(unittest.TestCase):
    
    def setUp(self):
        config = Config(NEO4J_URI)
        self.client = Neo4jClient(config)

    #def test_warm_cache(self):
    #    resp = self.client.warm_cache()
    #    print resp.raw

    def test_cypher(self):
        query = """START x  = node({_id}) MATCH x -[r]-> n RETURN type(r), n.name?, n.age?"""
        params = dict(_id=1261)
        resp = self.client.cypher(query,params)
        #print resp.raw

def neo4j_client_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(Neo4jClientTestCase))
    suite.addTest(unittest.makeSuite(Neo4jClientIndexTestCase))
    suite.addTest(unittest.makeSuite(Neo4jIndexTestCase))
    #suite.addTest(unittest.makeSuite(GremlinTestCase))
    #suite.addTest(unittest.makeSuite(CypherTestCase))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='neo4j_client_suite')




# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
import os
import sys
import inspect
import logging

import time
import datetime
import calendar
import omnijson as json # supports Python 2.5-3.2

#
# Python 3 
#

if sys.version < '3':
    import codecs
#    import ujson as json
    from urllib import quote, quote_plus, urlencode
    from urlparse import urlsplit, urlparse

    def u(x):
        return codecs.unicode_escape_decode(x)[0]
else:
    # ujson is faster but hasn't been ported to Python 3 yet
#    import json
    from urllib.parse import quote, quote_plus, urlencode, urlparse
    from urllib.parse import urlsplit

    long = int
    unicode = str

    def u(x):
        return x

#
# Logging
#

def get_logger(name,level=None):
    logger = logging.getLogger(name)
    if level:
        logger.setLevel(level)
    return logger

log = get_logger(__name__)

#
# Element Utils
#


def initialize_elements(client,response):
    # return None if there were no results; otherwise,
    # return a generator of initialized elements.
    if response.total_size > 0:
        # yield doesn't work for conditionals
        return (initialize_element(client, result) for result in response.results)

def initialize_element(client,result):
    # result should be a single Result object, not a list or generator
    element_class = get_element_class(client,result)
    element = element_class(client)
    element._initialize(result)
    return element

def get_element_class(client,result):
    element_key = get_element_key(client,result)
    element_class = client.registry.get_class(element_key)
    if element_class is None:
        # if element_class is not in registry, return the generic Vertex/Edge class
        base_type = result.get_type()
        element_class = client.registry.get_class(base_type)
    return element_class

def get_element_key(client,result):
    var_map = dict(vertex=client.config.type_var,
                   edge=client.config.label_var)
    base_type = result.get_type()
    if base_type == "vertex":
        key_var = var_map[base_type]
        # if key_var not found, just return the generic type for the Vertex
        element_key = result.data.get(key_var, base_type)
    elif base_type == "edge":
        label = result.get_label()
        element_key = label if label in client.registry.class_map else base_type
    else:
        raise TypeError
    return element_key

# Deprecated in favor of resp.one()
def get_one_result(resp):
    # If you're using this utility, that means the results attribute in the 
    # Response object should always contain a single result object,
    # not multiple items. But gremlin returns all results as a list
    # even if the list contains only one element. And the Response class
    # converts all lists to a generator of Result objects. Thus in that case,
    # we need to grab the single Result object out of the list/generator.
    if resp.total_size > 1:
        log.error('resp.results contains more than one item.')
        raise ValueError
    if inspect.isgenerator(resp.results):
        result = next(resp.results)
    else:
        result = resp.results
    return result
    

def get_key_value(key, value, pair):
    """Return the key and value, regardless of how it was entered."""
    if pair:
        key, value = pair.popitem()
    return key, value


#
# Client Utils
#

def build_path(*args):
    # don't include segment if it's None
    # quote_plus doesn't work for neo4j index lookups 
    # e.g., index/node/test_idxV/name/James+Thornton
    segments = [quote(unicode(segment), safe='') for segment in args if segment is not None]
    path = "/".join(segments)
    return path

#
# Time Utils
#

def current_timestamp():
    # Return the unix UTC time 
    # TODO: should we cast this to an int for consistency?
    return int(time.time())

def current_datetime():
    # Returns a UTC datetime object
    # return datetime.datetime.utcnow()
    now =  current_timestamp()
    #return datetime.datetime.utcfromtimestamp(now).replace(tzinfo=pytz.utc)
    return datetime.datetime.utcfromtimestamp(now)

def current_date():
    #Return  a date object
    return to_date(current_timestamp())
    
def to_timestamp(datetime):
    # Converts a datetime object to unix UTC time
    return calendar.timegm(datetime.utctimetuple()) 

def to_datestamp(date):
    # Converts a date object to unix UTC time
    return calendar.timegm(date.timetuple()) 

def to_datetime(timestamp):
    # Converts unix UTC time into a UTC datetime object
    #return datetime.datetime.utcfromtimestamp(timestamp).replace(tzinfo=pytz.utc)
    return datetime.datetime.utcfromtimestamp(timestamp)

def to_date(timestamp):
    # Converts unix UTC time into a date object
    return datetime.date.fromtimestamp(timestamp)
    
# Exaplanations on dealing with time...

    # http://unix4lyfe.org/time/
    # http://lucumr.pocoo.org/2011/7/15/eppur-si-muove/
    # http://nvie.com/posts/introducing-times/
    # http://news.ycombinator.com/item?id=3545935
    # http://labix.org/python-dateutil
    # http://docs.python.org/library/time.html#module-time
    # http://code.davidjanes.com/blog/2008/12/22/working-with-dates-times-and-timezones-in-python/

    # for historical dates, see:
    # http://www.egenix.com/products/python/mxBase/mxDateTime/

    # Always store UTC

    # One way (I think dateutils requires this)
    # t = time.time()  # unix utc timestamp
    # dt = datetime.datetime.utcfromtimestamp(t) 
    # ut = calendar.timegm(dt.utctimetuple()) 

    # Simpler way?
    # t = time.time()  # unix utc timestamp
    # dt = time.gmtime(t)
    # t = calendar.timegm(dt)

    # Both ways lose subsecond precision going from datetime object to unixtime    
    # t = time.mktime(dt)  # back to unix timestamp # don't use this, this is the inverse of localtime()



#
# Generic utils
#

def extract(desired_keys, bigdict):
    subset = dict([(i, bigdict[i]) for i in desired_keys if i in bigdict])
    return subset

def get_file_path(current_filename, target_filename):
    """
    Returns the full file path for the target file.
    
    """
    current_dir = os.path.dirname(current_filename)
    file_path = os.path.normpath(os.path.join(current_dir, target_filename))
    return file_path

def coerce_id(_id):
    """
    Tries to coerce a vertex ID into an integer and returns it.

    :param v: The vertex ID we want to coerce into an integer.
    :type v: int or str

    :rtype: int or str

    """
    try:
        return int(_id)
    except:
        # some DBs, such as OrientDB, use string IDs
        return _id







# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
from bulbs.config import Config
from bulbs.factory import Factory
from bulbs.element import Vertex, Edge
from bulbs.model import Relationship
from bulbs.utils import initialize_elements

from bulbs.base.client import Client
from bulbs.base.index import Index


# A framework is an understanding of how things could fit together.
# When designing these things, it's important to remember that your 
# understanding is incomplete.

# Bulbs is written as a series of layers, designed from the bottom up.

class Graph(object):
    """
    Abstract base class for the server-specific Graph implementations. 

    :param config: Optional Config object. Defaults to the default config.
    :type config: Config
        
    :cvar client_class: Client class.
    :cvar default_index: Default index class.

    :ivar client: Client object.
    :ivar config: Config object.
    :ivar vertices: VertexProxy object.
    :ivar edges: EdgeProxy object.

    Example:

    >>> from bulbs.neo4jserver import Graph
    >>> g = Graph()
    >>> james = g.vertices.create(name="James")
    >>> julie = g.vertices.create(name="Julie")
    >>> g.edges.create(james, "knows", julie)

    """    
    # The Client class to use for this Graph.
    client_class = Client

    # The default Index class.
    default_index = Index

    def __init__(self, config=None):
        self.client = self.client_class(config)
        self.config = self.client.config

        self.factory = Factory(self.client)

        self.vertices = self.build_proxy(Vertex)
        self.edges = self.build_proxy(Edge)

    @property
    def V(self):
        """
        Returns a list of all the vertices in the graph.

        :rtype: list or None

        """
        resp = self.client.get_all_vertices()
        if resp.total_size > 0:
            vertices = initialize_elements(self.client, resp)
            return list(vertices)
   
    @property
    def E(self):
        """
        Returns a list of all the edges in the graph.

        :rtype: list or None

        """
        resp = self.client.get_all_edges()
        if resp.total_size > 0:
            edges = initialize_elements(self.client, resp)
            return list(edges)
        
    def add_proxy(self, proxy_name, element_class, index_class=None):
        """
        Adds an element proxy to the Graph object for the element class.

        :param proxy_name: Attribute name to use for the proxy.
        :type proxy_name: str

        :param element_class: Element class managed by this proxy.
        :type element_class: Element

        :param index_class: Index class for Element's primary index. 
            Defaults to default_index.
        :type index_class: Index

        :rtype: None

        """
        proxy = self.build_proxy(element_class, index_class)
        self.client.registry.add_proxy(proxy_name, proxy)
        setattr(self, proxy_name, proxy)
    
    def build_proxy(self, element_class, index_class=None):
        """
        Returns an element proxy built to specifications.

        :param element_class: Element class managed by this proxy.
        :type element_class: Element

        :param index_class: Optional Index class for Element's primary index. 
            Defaults to default_index.
        :type index_class: Index

        :rtype: Element proxy

        """
        if not index_class:
            index_class = self.default_index
        return self.factory.build_element_proxy(element_class, index_class)

    def load_graphml(self, uri):
        """
        Loads a GraphML file into the database and returns the response.

        :param uri: URI of the GraphML file.
        :type uri: str

        :rtype: Response

        """
        raise NotImplementedError
        
    def get_graphml(self):
        """
        Returns a GraphML file representing the entire database.

        :rtype: Response

        """
        raise NotImplementedError

    def warm_cache(self):
        """
        Warms the server cache by loading elements into memory.

        :rtype: Response

        """
        raise NotImplementedError

    def clear(self):
        """Deletes all the elements in the graph.

        :rtype: Response

        .. admonition:: WARNING 

           This will delete all your data!

        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
from bulbs.element import Vertex, Edge


class Registry(object):
    """
    Store runtime configuration settings.

    :param config: Config object.
    :type config: bulbs.config.Config

    """
    
    def __init__(self, config):
        self.config = config
        self.class_map = dict(vertex=Vertex,edge=Edge)
        self.proxy_map = dict()
        self.index_map = dict()
        self.scripts_map = dict()

    # Classes

    def add_class(self, element_class):
        """
        Adds an element class to the registry.

        :param element_class: Element class.
        :type element_class: class

        :rtype: None

        """
        # Vertex and Edge are always set by default
        if element_class not in (Vertex, Edge): 
            # TODO: may get into an issue with name clashes
            # a vertex has the same element_type as an edge's label;
            # for now "don't do that".
            element_key = element_class.get_element_key(self.config)
            self.class_map[element_key] = element_class

    def get_class(self, element_key):
        """
        Returns the element class given the element key.

        :param element_key: Element key, value of element_type or label.
        :type element_key: str

        :rtype: class

        """
        return self.class_map.get(element_key)

    # Proxies

    def add_proxy(self, name, proxy):
        """
        Adds a proxy object to the registry.

        :param name: Proxy name.
        :type name: str

        :param proxy: Proxy object.
        :type proxy: object

        :rtype: None

        """
        self.proxy_map[name] = proxy

    def get_proxy(self, name):
        """
        Returns proxy objects given the name.

        :param name: Proxy name.
        :type name: str

        :rtype: class

        """
        return self.proxy_map[name]

    # Indices

    def add_index(self, index_name, index):
        """
        Adds an index object to the registry.

        :param name: Index name.
        :type name: str

        :param name: index
        :type name: Index

        :rtype: None

        """
        self.index_map[index_name] = index
        
    def get_index(self, index_name):
        """
        Returns the Index object for the given index name.

        :param index_name: Index name.
        :type index_name: str

        :rtype: Index

        """
        return self.index_map[index_name]
 
    # Scripts

    def add_scripts(self, name, scripts):
        """
        Adds a scripts object to the registry.

        :param name: Scripts object name.
        :type name: str

        :param name: Scripts object.
        :type name: Scripts

        :rtype: None

        """
        self.scripts_map[name] = scripts

    def get_scripts(self, key):
        """
        Returns a scripts object for the given name.

        :param name: Scripts object name.
        :type name: str

        :rtype: Scripts

        """
        return self.scripts_map[key]




from .client import Neo4jRequest, Neo4jClient

#
# Batch isn't fully baked yet
#

class Neo4jBatchRequest(Neo4jRequest):
    """Makes HTTP requests to Neo4j Server and returns a Neo4jResponse.""" 
    
    def _initialize(self):
        self.messages = []
        self.message_id = 0

    def request(self, method, path, params):
        """
        Adds request to the messages list and returns a placeholder.

        :param method: HTTP method: GET, PUT, POST, or DELETE.
        :type method: str

        :param path: Path to the server resource, relative to the root URI.
        :type path: str

        :param params: Optional URI parameters for the resource.
        :type params: dict

        :rtype: str

        """
                
        return self.add_message(method, path, params)

        # return self 
        # would allow you to do self.request.post(path, params).send()
        # that won't work unless you always want to go throught the batch interface


    def add_message(self, method, path, params):
        message_id = self.next_id()
        message = dict(method=method, to=path, body=params, id=message_id)
        self.messages.append(message)
        return self.placeholder(message_id)

    def next_id(self):
        self.message_id = self.message_id + 1
        return self.message_id

    def placeholder(self, message_id):
        return "{%d}" % message_id

    def send(self):
        """
        Convenience method that sends request messages to the client.

        :param message: Tuple containing: (HTTP method, path, params)
        :type path: tuple

        :param params: Optional URI params for the resource.
        :type params: dict

        :rtype: Response

        """
        path = "batch"
        params = self.messages
        return Neo4jRequest.post(self, path, params)

    def get_messages(self):
        return self.messages

    def clear(self):
        self._initialize()


class Neo4jBatchClient(Neo4jClient):

    request_class = Neo4jBatchRequest

    # Batch isn't fully baked yet

    # Batch try (old -- from Neo4jClient)...
    #def create_indexed_vertex(self,data,index_name,keys=None):
    #    """Creates a vertex, indexes it, and returns the Response."""
    #    batch = Neo4jBatch(self.client)
    #    placeholder = batch.add(self.message.create_vertex(data))
    #    for key in keys:
    #        value = data.get(key)
    #        if value is None: continue
    #        batch.add(self.message.put_vertex(index_name,key,value,placeholder))
    #    resp = batch.send()
    #    #for result in resp.results:


    def send(self):
        return self.request.send()

    def get_messages(self):
        return self.request.get_messages()

    def clear(self):
        self.request.clear()





# -*- coding: utf-8 -*-
#
# Copyright 2011 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
import unittest

from bulbs.tests.testcase import BulbsTestCase
from bulbs.element import Vertex, VertexProxy, Edge, EdgeProxy
from bulbs.config import Config
    
from bulbs.rexster import RexsterClient, REXSTER_URI
from bulbs.rexster.index import VertexIndexProxy, EdgeIndexProxy, ManualIndex

config = Config(REXSTER_URI)
BulbsTestCase.client = RexsterClient(config)
BulbsTestCase.index_class = ManualIndex

 
class IndexTestCase(BulbsTestCase):
    
    def setUp(self):
        self.indicesV = VertexIndexProxy(self.index_class,self.client)
        self.indicesE = EdgeIndexProxy(self.index_class,self.client)

        self.indicesV.delete("test_idxV")
        self.indicesE.delete("test_idxE")

        self.vertices = VertexProxy(Vertex,self.client)
        self.vertices.index = self.indicesV.get_or_create("test_idxV")

        self.edges = EdgeProxy(Edge,self.client)
        self.edges.index = self.indicesE.get_or_create("test_idxE")

    def test_index(self):
        index_name = "test_idxV"
        # need to fix this to accept actual data types in POST
        #ikeys = '[name,location]'
        #self.indices.delete(index_name)
        #i1 = self.indices.create(index_name,Vertex)
        #assert i1.index_name == index_name
        #assert i1.index_type == "automatic"

        james = self.vertices.create({'name':'James'})
        self.vertices.index.put(james._id,'name','James')
        self.vertices.index.put(james._id,'location','Dallas')
        results = self.vertices.index.lookup('name','James')
        results = list(results)
        #print "RESULTS", results
        assert len(results) == 1
        assert results[0].name == "James"
        total_size = self.vertices.index.count('name','James')
        assert total_size == 1
        # NOTE: only automatic indices have user provided keys
        #keys = self.vertices.index..keys()
        #assert 'name' in keys
        #assert 'location' in keys
        i2 = self.indicesV.get(index_name)
        #print "INDEX_NAME", index_name, self.vertices.index..index_name, i2.index_name
        assert self.vertices.index.index_name == i2.index_name
        
        # remove vertex is bugged
        #self.vertices.index..remove(james._id,'name','James')
        #james = self.vertices.index..get_unique('name','James')
        #assert james is None
  
        # only can rebuild automatic indices
        #i3 = self.indices.get("vertices",Vertex)
        #results = i3.rebuild()
        #assert type(results) == list

        self.indicesV.delete(index_name)

        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(IndexTestCase))

    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='suite')




# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
from bulbs.config import Config, DEBUG, INFO, WARNING, ERROR, CRITICAL
from bulbs.element import Vertex, VertexProxy, Edge, EdgeProxy
from bulbs.model import Node, NodeProxy, Relationship, RelationshipProxy

from .graph import Graph
#from .batch import Neo4jBatch
from .client import Neo4jClient, NEO4J_URI
from .index import ExactIndex, FulltextIndex, AutomaticIndex, \
    VertexIndexProxy, EdgeIndexProxy



# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""

Bulbs supports pluggable backends. These are the abstract base classes that 
provides the index interface for a client. Implement these to create an index.

"""
from bulbs.utils import initialize_element, initialize_elements, get_one_result


# Index Proxies

class VertexIndexProxy(object):
    """
    Abstract base class the vertex index proxy.

    :ivar index_class: Index class.
    :ivar client: Client object.

    """
    def __init__(self, index_class, client):        
        # The index class for this proxy, e.g. ExactIndex.
        self.index_class = index_class

        # The Client object for the database.
        self.client = client
    
    def create(self, index_name):
        """
        Creates an Vertex index and returns it.

        :param index_name: Index name.
        :type index_name: str

        :rtype: Index
        
        """
        raise NotImplementedError

    def get(self, index_name):
        """
        Returns the Index object with the specified name or None if not found.
        
        :param index_name: Index name.
        :type index_name: str

        :rtype: Index
        
        """
        raise NotImplementedError

    def get_or_create(self, index_name):
        """
        Get a Vertex Index or create it if it doesn't exist.

        :param index_name: Index name.
        :type index_name: str

        :rtype: Index

        """ 
        raise NotImplementedError

    def delete(self, index_name):
        """ 
        Deletes an index and returns the Response.

        :param index_name: Index name.
        :type index_name: str

        :rtype: Response

        """
        raise NotImplementedError


class EdgeIndexProxy(object):
    """
    Abstract base class the edge index proxy.

    :ivar index_class: Index class.
    :ivar client: Client object.

    """
    def __init__(self, index_class, client):        
        # The index class for this proxy, e.g. ExactIndex.
        self.index_class = index_class

        # The Client object for the database.
        self.client = client
    
    def create(self, index_name):
        """
        Creates an Edge index and returns it.

        :param index_name: Index name.
        :type index_name: str

        :rtype: Index
        
        """
        raise NotImplementedError

    def get(self, index_name):
        """
        Returns the Index object with the specified name or None if not found.
        
        :param index_name: Index name.
        :type index_name: str

        :rtype: Index
        
        """
        raise NotImplementedError

    def get_or_create(self, index_name):
        """
        Get an Edge Index or create it if it doesn't exist.

        :param index_name: Index name.
        :type index_name: str

        :rtype: Index

        """ 
        raise NotImplementedError

    def delete(self, index_name):
        """ 
        Deletes an index and returns the Response.

        :param index_name: Index name.
        :type index_name: str

        :rtype: Response

        """
        raise NotImplementedError


# Index Containers

class Index(object):
    """
    Abstract base class for the default index.

    :ivar client: Client object 
    :ivar result: Result object.

    """

    def __init__(self, client, result):

        # The Client object for the database.
        self.client = client

        # The index attributes returned by the proxy request.
        self.result = result

    @classmethod 
    def get_proxy_class(cls, base_type=None):
        """
        Returns the IndexProxy class.

        :param base_type: Index base type, either vertex or edge.
        :type base_type: str

        :rtype: class

        """
        class_map = dict(vertex=VertexIndexProxy, edge=EdgeIndexProxy)
        return class_map[base_type]

    @property
    def index_name(self):
        """
        Returns the index name.

        :rtype: str

        """
        return self.result.get_index_name()

    @property
    def index_class(self):
        """
        Returns the index class.

        :rtype: class

        """
        return self.result.get_index_class()

    def put(self, _id, key=None, value=None, **pair):
        """
        Put an element into the index at key/value and return the Response.

        :param _id: The element ID.
        :type _id: int or str

        :param key: The index key. 
        :type key: str

        :param value: The key's value.
        :type value: str or int

        :param pair: Optional key/value pair. Example: name="James"
        :type pair: name/value pair

        :rtype: Response
            
        """
        raise NotImplementedError

    def update(self, _id, key=None, value=None, **pair):
        """
        Update the element ID for the key and value.
        
        :param key: The index key. 
        :type key: str

        :param value: The key's value.
        :type value: str or int

        :param pair: Optional key/value pair. Example: name="James"
        :type pair: key/value pair

        :rtype: Response

        """
        raise NotImplementedError

    def lookup(self, key=None, value=None, **pair):
        """
        Return all the elements in the index where key equals value.

        :param key: The index key. 
        :type key: str

        :param value: The key's value.
        :type value: str or int

        :param pair: Optional key/value pair. Example: name="James"
        :type pair: key/value pair

        :rtype: Element generator

        """
        raise NotImplementedError

    def put_unique(self, _id, key=None, value=None, **pair):
        """
        Put an element into the index at key/value and overwrite it if an 
        element already exists; thus, when you do a lookup on that key/value pair,
        there will be a max of 1 element returned.

        :param key: The index key. 
        :type key: str

        :param value: The key's value.
        :type value: str or int

        :param pair: Optional key/value pair. Example: name="James"
        :type pair: key/value pair

        :rtype: Resposne

        """
        raise NotImplementedError

    def get_unique(self, key=None, value=None, **pair):
        """
        Returns a max of 1 elements in the index matching the key/value pair.

        :param key: The index key. 
        :type key: str

        :param value: The key's value.
        :type value: str or int

        :param pair: Optional key/value pair. Example: name="James"
        :type pair: key/value pair

        :rtype: Element or None

        """
        raise NotImplementedError

    def remove(self, _id, key=None, value=None, **pair):
        """
        Remove the element from the index located by key/value.

        :param key: The index key. 
        :type key: str

        :param value: The key's value.
        :type value: str or int

        :param pair: Optional key/value pair. Example: name="James"
        :type pair: key/value pair

        :rtype: Response

        """
        raise NotImplementedError

    def count(self, key=None, value=None, **pair):
        """
        Return the number of items in the index for the key and value.

        :param key: The index key. 
        :type key: str

        :param value: The key's value.
        :type value: str or int

        :param pair: Optional key/value pair. Example: name="James"
        :type pair: name/value pair

        :rtype: int

        """
        raise NotImplementedError
        
    def _get_key_value(self, key, value, pair):
        """
        Returns the key and value, regardless of how it was entered.

        :param key: The index key. 
        :type key: str

        :param value: The key's value.
        :type value: str or int

        :param pair: Optional key/value pair. Example: name="James"
        :type pair: key/value pair

        :rtype: tuple

        """
        if pair:
            key, value = pair.popitem()
        return key, value

    def _get_method(self, **method_map):
        """
        Returns the right method, depending on the index class type.

        :param method_map: Dict mapping the index class type to its method name. 
        :type method_map: dict

        :rtype: Callable

        """
        method_name = method_map[self.index_class]
        method = getattr(self.client, method_name)
        return method



"""
Bulbs
-----

Bulbs is a Python persistence framework for graph databases that 
connects to Neo4j Server, Rexster, OrientDB, Lightsocket, and more.

"""
import sys
from setuptools import Command, setup, find_packages

class run_audit(Command):
    """Audits source code using PyFlakes for following issues:
       - Names which are used but not defined or used before they are defined.
       - Names which are redefined without having been used.
    """
    description = "Audit source code with PyFlakes"
    user_options = []

    def initialize_options(self):
        all = None

    def finalize_options(self):
        pass

    def run(self):
        import os, sys
        try:
            import pyflakes.scripts.pyflakes as flakes
        except ImportError:
            print("Audit requires PyFlakes installed in your system.")
            sys.exit(-1)

        dirs = ['bulbs', 'tests']
        # Add example directories
        #for dir in ['blog',]:
        #    dirs.append(os.path.join('examples', dir))
        # TODO: Add test subdirectories
        warns = 0
        for dir in dirs:
            for filename in os.listdir(dir):
                if filename.endswith('.py') and filename != '__init__.py':
                    warns += flakes.checkPath(os.path.join(dir, filename))
        if warns > 0:
            print(("Audit finished with total %d warnings." % warns))
        else:
            print("No problems found in sourcecode.")

def run_tests():
    import os, sys
    sys.path.append(os.path.join(os.path.dirname(__file__), 'tests'))
    from bulbs_tests import suite
    return suite()

# Python 3
install_requires = ['distribute', 'httplib2>=0.7.2', 'pyyaml>=3.10', 'six', 'omnijson']
if sys.version < '3':
    install_requires.append('python-dateutil==1.5')
else:
    # argparse is in 3.2 but not 3.1
    install_requires.append('argparse')
    install_requires.append('python-dateutil>=2')


setup (
    name = 'bulbs',
    version = '0.3.14',
    url = 'https://github.com/espeed/bulbs',
    license = 'BSD',
    author = 'James Thornton',
    author_email = 'james@jamesthornton.com',
    description = 'A Python persistence framework for graph databases that '
                  'connects to Neo4j Server, Rexster, OrientDB, Lightsocket.',
    long_description = __doc__,
    keywords = "graph database DB persistence framework rexster gremlin cypher neo4j orientdb",   
    packages = find_packages(),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=install_requires, 
    classifiers = [
        "Programming Language :: Python",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2.6',
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Database",
        "Topic :: Database :: Front-Ends",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Distributed Computing",
        ],
    cmdclass={'audit': run_audit},
    test_suite='__main__.run_tests'
)



# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
    Bulbs is a Python persistence framework for graph databases.
    
    Bulbs is a Python client library for Neo4j Server and Rexster, 
    Rexster is a REST server that provides access to any 
    Blueprints-enabled graph database, including Neo4j, OrientDB, DEX, 
    TinkerGraph and OpenRDF.

    Home Page: http://bulbflow.com
 
    Copyright (c) 2012, James Thornton (http://jamesthornton.com)
    License: BSD (see LICENSE for details)
"""
__version__ = '0.3'
__author__ = 'James Thornton'
__license__ = 'BSD'








import unittest
import random

from bulbs.config import Config, DEBUG, ERROR
from bulbs.registry import Registry
from bulbs.base import TypeSystem

class ClientIndexTestCase(unittest.TestCase):
    
    def setUp(self):
        self.client = None
        raise NotImplementedError

    # Some server implementations (Rexster) return a 404 if index doesn't exist
    def _delete_vertex_index(self,index_name):
        try:
            self.client.delete_vertex_index(index_name)
        except LookupError:
            pass

    def _delete_edge_index(self,index_name):
        try:
            self.client.delete_edge_index(index_name)
        except LookupError:
            pass

    # Index Controller Tests

    def test_create_vertex_index(self):
        index_name = "test_idxV"
        self._delete_vertex_index(index_name)
        resp = self.client.create_vertex_index(index_name)

        assert resp.results.get_index_class() == "vertex"        
        assert resp.results.get_index_name() == index_name
                
    def test_get_vertex_index(self):
        index_name = "test_idxV"
        resp = self.client.get_vertex_index(index_name)

        assert resp.results.get_index_class() == "vertex"        
        assert resp.results.get_index_name() == index_name


    # Index Container Tests

    def test_indexed_vertex_CRUD(self):

        index_name = "test_idxV"
        self._delete_vertex_index(index_name)
        self.client.create_vertex_index(index_name)

        # Create and Index Vertex
        name1 = "James %s" % random.random()
        age1 = 34
        data1 = dict(name=name1, age=age1)
        keys1 = ['name']
        self.client.create_indexed_vertex(data1, index_name, keys1)
        
        # Lookup Vertex
        resp1 = self.client.lookup_vertex(index_name, "name", name1)
        results1 = next(resp1.results)

        assert results1.get_type() == "vertex"
        assert results1.get_data() == data1  
        assert results1.data.get('name') == name1
        assert results1.data.get('age') == age1

        # Update and Index Vertex (update doesn't return data)
        _id = results1.get_id()
        name2 = "James Thornton %s" % random.random()
        age2 = 35
        data2 = dict(name=name2, age=age2)
        keys2 = None
        self.client.update_indexed_vertex(_id, data2, index_name, keys2)

        # Lookup Vertex
        resp2 = self.client.lookup_vertex(index_name, "name", name2)
        result2 = next(resp2.results)

        assert result2.get_type() == "vertex"
        assert result2.get_data() == data2
        assert result2.data.get('name') == name2
        assert result2.data.get('age') == age2

        # Remove a vertex from the index
        self.client.remove_vertex(index_name, _id, "name", name2)
        resp3 = self.client.lookup_vertex(index_name, "name", name2)
        assert resp3.total_size == 0
        
    def test_indexed_edge_CRUD(self):
        index_name = "test_idxE"
        self._delete_edge_index(index_name)
        self.client.create_edge_index(index_name)

        respV1 = self.client.create_vertex({'name':'James','age':34})
        respV2 = self.client.create_vertex({'name':'Julie','age':28})
        V1_id = respV1.results.get_id()
        V2_id = respV2.results.get_id()

        # Create and Index Edge
        city1 = "Dallas"
        data1 = dict(city=city1)
        keys1 = ['city']
        resp = self.client.create_indexed_edge(V1_id, "knows", V2_id, data1, index_name, keys1)
        
        # Lookup Edge
        resp1 = self.client.lookup_edge(index_name, "city", city1)
        results1 = next(resp1.results)

        assert results1.get_type() == "edge"
        assert results1.get_data() == data1  
        assert results1.data.get('city') == city1

        # Update and Index Edge (update doesn't return data)
        _id = results1.get_id()
        city2 = "Austin"
        data2 = dict(city=city2)
        keys2 = ['city']
        self.client.update_indexed_edge(_id, data2, index_name, keys2)

        # Lookup Edge
        resp2 = self.client.lookup_edge(index_name, "city", city2)
        result2 = next(resp2.results)

        assert result2.get_type() == "edge"
        assert result2.get_data() == data2
        assert result2.data.get('city') == city2

        # Remove and edge from the index
        self.client.remove_edge(index_name, _id, "city", city2)
        
        resp3 = self.client.lookup_edge(index_name, "city", city2)
        assert resp3.total_size == 0
        





# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
An interface for interacting with indices on Rexster.

"""
from bulbs.utils import initialize_element, initialize_elements, get_one_result


class IndexProxy(object):
    """Abstract base class the index proxies."""

    def __init__(self, index_class, client):        
        # The index class for this proxy, e.g. ManualIndex.
        self.index_class = index_class

        # The Client object for the database.
        self.client = client
    

class VertexIndexProxy(IndexProxy):
    """
    Manage vertex indices on Rexster.

    :param index_class: The index class for this proxy, e.g. ManualIndex.
    :type index_class: Index

    :param client: The Client object for the database.
    :type client: bulbs.rexster.client.RexsterClient

    :ivar index_class: Index class.
    :ivar client: RexsterClient object.

    """
                        
    def create(self, index_name):
        """
        Creates an Vertex index and returns it.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index
        
        """
        resp = self.client.create_vertex_index(index_name)
        index = self.index_class(self.client,resp.results)
        self.client.registry.add_index(index_name, index)
        return index

    def get(self, index_name):
        """
        Returns the Index object with the specified name or None if not found.
        
        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index
        
        """
        resp = self.client.get_vertex_index(index_name)
        if resp.results:
            index = self.index_class(self.client,resp.results)
            self.client.registry.add_index(index_name, index)
            return index

    def get_or_create(self, index_name, index_params=None):
        """
        Get a Vertex Index or create it if it doesn't exist.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index

        """ 
        resp = self.client.get_or_create_vertex_index(index_name, index_params)
        index = self.index_class(self.client,resp.results)
        self.client.registry.add_index(index_name, index)
        return index

    def delete(self, index_name):
        """ 
        Deletes an index and returns the Response.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.client.RexsterResponse

        """
        try:
            return self.client.delete_vertex_index(index_name)
        except LookupError:
            return None


class EdgeIndexProxy(IndexProxy):
    """
    Manage edge indices on Rexster.

    :param index_class: The index class for this proxy, e.g. ManualIndex.
    :type index_class: Index

    :param client: The Client object for the database.
    :type client: bulbs.rexster.client.RexsterClient

    :ivar index_class: Index class.
    :ivar client: RexsterClient object.

    """

    def create(self,index_name,*args,**kwds):
        """ 
        Adds an index to the database and returns it. 

        index_keys must be a string in this format: '[k1,k2]'
        Don't pass actual list b/c keys get double quoted.

        :param index_name: The name of the index to create.

        :param index_class: The class of the elements stored in the index. 
                            Either vertex or edge.
        
        """
        resp = self.client.create_edge_index(index_name,*args,**kwds)
        index = self.index_class(self.client,resp.results)
        self.client.registry.add_index(index_name,index)
        return index

    def get(self,index_name):
        """
        Returns the Index object with the specified name or None if not found.
        
        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index
        
        """
        resp = self.client.get_edge_index(index_name)
        if resp.results:
            index = self.index_class(self.client,resp.results)
            self.client.registry.add_index(index_name,index)
            return index

    def get_or_create(self, index_name, index_params=None):
        """
        Get an Edge Index or create it if it doesn't exist.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.index.Index

        """ 
        resp = self.client.get_or_create_edge_index(index_name, index_params)
        index = self.index_class(self.client,resp.results)
        self.client.registry.add_index(index_name, index)
        return index

    def delete(self,index_name):
        """ 
        Deletes an index and returns the Response.

        :param index_name: Index name.
        :type index_name: str

        :rtype: bulbs.rexster.client.RexsterResponse

        """
        try:
            return self.client.delete_edge_index(index_name)
        except LookupError:
            return None


#
# Index Containers (Manual, Automatic)
#

class Index(object):
    """Abstract base class for an index."""

    def __init__(self, client, result):
        self.client = client
        self.result = result

    @classmethod 
    def get_proxy_class(cls, base_type):
        """
        Returns the IndexProxy class.

        :param base_type: Index base type, either vertex or edge.
        :type base_type: str

        :rtype: class

        """
        class_map = dict(vertex=VertexIndexProxy, edge=EdgeIndexProxy)
        return class_map[base_type]

    @property
    def index_name(self):
        """
        Returns the index name.

        :rtype: str

        """
        return self.result.data['name']

    @property
    def index_class(self):
        """
        Returns the index class, either vertex or edge.

        :rtype: class

        """
        return self.result.data['class']

    @property
    def index_type(self):
        """
        Returns the index type, which will either be automatic or manual.

        :rtype: str

        """
        return self.result.data['type']

    def count(self,key=None,value=None,**pair):
        """
        Return a count of all elements with 'key' equal to 'value' in the index.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param pair: Optional keyword param. Instead of supplying key=name 
                     and value = 'James', you can supply a key/value pair in
                     the form of name='James'.
        """
        key, value = self._get_key_value(key,value,pair)
        resp = self.client.index_count(self.index_name,key,value)
        return resp.content['totalSize']


    def _get_key_value(self, key, value, pair):
        """Return the key and value, regardless of how it was entered."""
        if pair:
            key, value = pair.popitem()
        return key, value

    def _get_method(self, **method_map):
        method_name = method_map[self.index_class]
        method = getattr(self.client, method_name)
        return method

    def lookup(self, key=None, value=None, **pair):
        """
        Return a generator containing all the elements with key property equal 
        to value in the index.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param raw: Optional keyword param. If set to True, it won't try to 
                    initialize the results. Defaults to False. 

        :param pair: Optional keyword param. Instead of supplying key=name 
                     and value = 'James', you can supply a key/value pair in
                     the form of name='James'.
        """
        key, value = self._get_key_value(key, value, pair)
        resp = self.client.lookup_vertex(self.index_name,key,value)
        return initialize_elements(self.client,resp)


class ManualIndex(Index):
    """
    Creates, retrieves, and deletes indices provided by the graph database.

    Use this class to get, put, and update items in an index.
    

    :param client: The Client object for the database.

    :param result: The result list returned by Rexster.

    :param classes: Zero or more subclasses of Element to use when 
                    initializing the the elements returned by the query. 
                    For example, if Person is a subclass of Node (which 
                    is defined in model.py and is a subclass of Vertex), 
                    and the query returns person elements, pass in the 
                    Person class and the method will use the element_type
                    defined in the class to initialize the returned items
                    to a Person object.

    Example that creates an index for Web page URL stubs, 
    adds an page element to it, and then retrieves it from the index::

        >>> graph = Graph()
        >>> graph.indices.create("page","vertex","automatic","[stub]")
        >>> index = graph.indices.get("page")
        >>> index.put("stub",stub,page._id)
        >>> page = index.get("stub",stub)

    """

    
    def put(self,_id,key=None,value=None,**pair):
        """
        Put an element into the index at key/value and return Rexster's 
        response.

        :param _id: The element ID.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param pair: Optional keyword param. Instead of supplying key=name 
                     and value = 'James', you can supply a key/value pair in
                     the form of name='James'.

        """
        # NOTE: if you ever change the _id arg to element, change remove() too
        key, value = self._get_key_value(key,value,pair)
        put = self._get_method(vertex="put_vertex", edge="put_edge")
        resp = put(self.index_name,key,value,_id)
        return resp

    def update(self,_id,key=None,value=None,**pair):
        """
        Update the element ID for the key and value and return Rexsters' 
        response.

        :param _id: The element ID.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param pair: Optional keyword param. Instead of supplying key=name 
                     and value = 'James', you can supply a key/value pair in
                     the form of name='James'.
        """
        key, value = self._get_key_value(key,value,pair)
        for result in self.get(key,value):
            self.remove(self.index_name, result._id, key, value)
        return self.put(_id,key,value)


    def put_unique(self,_id,key=None,value=None,**pair):
        """
        Put an element into the index at key/value and overwrite it if an 
        element already exists at that key and value; thus, there will be a max
        of 1 element returned for that key/value pair. Return Rexster's 
        response.

        :param _id: The element ID.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param pair: Optional keyword param. Instead of supplying key=name 
                     and value = 'James', you can supply a key/value pair in 
                     the form of name='James'.

        """
        return self.update(_id, key, value, **pair)


    def get_unique(self,key=None,value=None,**pair):
        """
        Returns a max of 1 elements matching the key/value pair in the index.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param pair: Optional keyword param. Instead of supplying key=name 
                     and value = 'James', you can supply a key/value pair in
                     the form of name='James'.
        """
        key, value = self._get_key_value(key,value,pair)
        resp = self.client.lookup_vertex(self.index_name,key,value)
        if resp.total_size > 0:
            result = get_one_result(resp)
            return initialize_element(self.client, result)

    def remove(self,_id,key=None,value=None,**pair):
        """
        Remove the element from the index located by key/value.

        :param _id: The element ID.

        :param key: The index key. This is optional because you can instead 
                    supply a key/value pair such as name="James". 

        :param value: The index key's value. This is optional because you can 
                      instead supply a key/value pair such as name="James". 

        :param pair: Optional keyword param. Instead of supplying key=name 
                     and value = 'James', you can supply a key/value pair in
                     the form of name='James'.
        """
        key, value = self._get_key_value(key, value, pair)
        remove = self._get_method(vertex="remove_vertex", edge="remove_edge")
        return remove(self.index_name,_id,key,value)


class AutomaticIndex(Index):

    def keys(self):
        """Return the index's keys."""
        resp = self.client.index_keys(self.index_name)
        return list(resp.results)
    
    def rebuild(self):
        # need class_map b/c the Blueprints need capitalized class names, 
        # but Rexster returns lower-case class names for index_class
        method_map = dict(vertex=self.client.rebuild_vertex_index,
                          edge=self.client.rebuild_edge_index)
        rebuild_method = method_map.get(self.index_class)
        resp = rebuild_method(self.index_name)
        return list(resp.results)

 



# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Bulbs supports pluggable clients. This is the Rexster client.

"""
from bulbs.config import Config, DEBUG
from bulbs.registry import Registry
from bulbs.utils import get_logger

# specific to this client
from bulbs.json import JSONTypeSystem
from bulbs.base import Client, Response, Result 
from bulbs.rest import Request, RESPONSE_HANDLERS
from bulbs.groovy import GroovyScripts

from bulbs.utils import json, build_path, get_file_path, urlsplit, coerce_id


# The default URIs
REXSTER_URI = "http://localhost:8182/graphs/tinkergraph"
SAIL_URI = "http://localhost:8182/graphs/sailgraph"

# The logger defined in Config
log = get_logger(__name__)

# Rexster resource paths
# TODO: local path vars would be faster
vertex_path = "vertices"
edge_path = "edges"
index_path = "indices"
gremlin_path = "tp/gremlin"
transaction_path = "tp/batch/tx"
multi_get_path = "tp/batch"


class RexsterResult(Result):
    """
    Container class for a single result, not a list of results.

    :param result: The raw result.
    :type result: dict

    :param config: The client Config object.
    :type config: Config 

    :ivar raw: The raw result.
    :ivar data: The data in the result.

    """
    def __init__(self, result, config):
        self.config = config

        # The raw result.
        self.raw = result

        # The data in the result.
        self.data = result

    def get_id(self):
        """
        Returns the element ID.

        :rtype: int or str

        """
        _id = self.data['_id']

        # OrientDB uses string IDs
        return coerce_id(_id)
               
    def get_type(self):
        """
        Returns the element's base type, either "vertex" or "edge".

        :rtype: str

        """
        return self.data['_type']
        
    def get_data(self):
        """
        Returns the element's property data.

        :rtype: dict

        """
        property_data = dict()
        private_keys = ['_id','_type','_outV','_inV','_label']
        for key in self.data: # Python 3
            value = self.data[key]
            if key not in private_keys:
                property_data.update({key:value})
        return property_data

    def get_uri(self):
        """
        Returns the element URI.

        :rtype: str

        """
        path_map = dict(vertex="vertices",edge="edges")
        _id = self.get_id()
        _type = self.get_type()
        element_path = path_map[_type]
        root_uri = self.config.root_uri
        uri = "%s/%s/%s" % (root_uri,element_path,_id)
        return uri
                 
    def get_outV(self):
        """
        Returns the ID of the edge's outgoing vertex (start node).

        :rtype: int

        """
        _outV = self.data.get('_outV')
        return coerce_id(_outV)
        
    def get_inV(self):
        """
        Returns the ID of the edge's incoming vertex (end node).

        :rtype: int

        """
        _inV = self.data.get('_inV')
        return coerce_id(_inV)

    def get_label(self):
        """
        Returns the edge label (relationship type).

        :rtype: str

        """
        return self.data.get('_label')

    def get_index_name(self):
        """
        Returns the index name.

        :rtype: str

        """
        return self.data['name']

    def get_index_class(self):
        """
        Returns the index class, either "vertex" or "edge".

        :rtype: str

        """
        return self.data['class']

    def get(self,attribute):
        """
        Returns the value of a client-specific attribute.
        
        :param attribute: Name of the attribute. 
        :type attribute: str

        :rtype: str

        """
        return self.data[attribute]


class RexsterResponse(Response):
    """
    Container class for the server response.

    :param response: httplib2 response: (headers, content).
    :type response: tuple

    :param config: Config object.
    :type config: bulbs.config.Config

    :ivar config: Config object.
    :ivar headers: httplib2 response headers, see:
        http://httplib2.googlecode.com/hg/doc/html/libhttplib2.html
    :ivar content: A dict containing the HTTP response content.
    :ivar results: A generator of RexsterResult objects, a single RexsterResult object, 
        or None, depending on the number of results returned.
    :ivar total_size: The number of results returned.
    :ivar raw: Raw HTTP response. Only set when log_level is DEBUG.

    """
    result_class = RexsterResult

    def __init__(self, response, config):
        self.config = config
        self.handle_response(response)
        self.headers = self.get_headers(response)
        self.content = self.get_content(response)
        self.results, self.total_size = self.get_results()
        self.raw = self._maybe_get_raw(response, config)

    def _maybe_get_raw(self,response, config):
        """Returns the raw response if in DEBUG mode."""
        # don't store raw response in production else you'll bloat the obj
        if config.log_level == DEBUG:
            return response

    def handle_response(self,http_resp):
        """
        Check the server response and raise exception if needed.
        
        :param response: httplib2 response: (headers, content).
        :type response: tuple

        :rtype: None

        """
        headers, content = http_resp
        response_handler = RESPONSE_HANDLERS.get(headers.status)
        response_handler(http_resp)

    def get_headers(self,response):
        """
        Returns a dict containing the headers from the response.

        :param response: httplib2 response: (headers, content).
        :type response: tuple
        
        :rtype: httplib2.Response

        """
        headers, content = response
        return headers

    def get_content(self,response):
        """
        Returns a dict containing the content from the response.
        
        :param response: httplib2 response: (headers, content).
        :type response: tuple
        
        :rtype: dict or None

        """
        # response is a tuple containing (headers, content)
        # headers is an httplib2 Response object, content is a string
        # see http://httplib2.googlecode.com/hg/doc/html/libhttplib2.html
        headers, content = response

        if content:
            content = json.loads(content.decode('utf-8'))
            return content

    def get_results(self):
        """
        Returns the results contained in the response.

        :return:  A tuple containing two items: 1. Either a generator of RexsterResult objects, 
                  a single RexsterResult object, or None, depending on the number of results 
                  returned; 2. An int representing the number results returned.
        :rtype: tuple

        """
        if type(self.content.get('results')) == list:
            results = (self.result_class(result, self.config) for result in self.content['results'])
            total_size = len(self.content['results'])
        elif self.content.get('results'):
            results = self.result_class(self.content['results'], self.config)
            total_size = 1
        else:
            results = None
            total_size = 0
        return results, total_size


class RexsterRequest(Request):
    """Makes HTTP requests to Rexster and returns a RexsterResponse.""" 
    
    response_class = RexsterResponse


class RexsterClient(Client):
    """
    Low-level client that sends a request to Rexster and returns a response.

    :param config: Optional Config object. Defaults to default Config.
    :type config: bulbs.config.Config

    :cvar default_uri: Default URI for the database.
    :cvar request_class: Request class for the Client.

    :ivar config: Config object.
    :ivar registry: Registry object.
    :ivar scripts: GroovyScripts object.  
    :ivar type_system: JSONTypeSystem object.
    :ivar request: RexsterRequest object.

    Example:

    >>> from bulbs.rexster import RexsterClient
    >>> client = RexsterClient()
    >>> script = client.scripts.get("get_vertices")
    >>> response = client.gremlin(script, params=None)
    >>> result = response.results.next()

    """ 
    #: Default URI for the database.
    default_uri = REXSTER_URI
    request_class = RexsterRequest


    def __init__(self, config=None, db_name=None):
        # This makes is easy to test different DBs 
        uri = self._get_uri(db_name) or self.default_uri

        self.config = config or Config(uri)
        self.registry = Registry(self.config)
        self.type_system = JSONTypeSystem()
        self.request = self.request_class(self.config, self.type_system.content_type)

        # Rexster supports Gremlin so include the Gremlin-Groovy script library
        self.scripts = GroovyScripts() 

        # Also include the Rexster-specific Gremlin-Groovy scripts
        scripts_file = get_file_path(__file__, "gremlin.groovy")
        self.scripts.update(scripts_file)

        # Add it to the registry. This allows you to have more than one scripts namespace.
        self.registry.add_scripts("gremlin", self.scripts)

    def _get_uri(self, db_name):
        if db_name is not None:
            uri = "http://localhost:8182/graphs/%s" % db_name
            return uri

    # Gremlin

    def gremlin(self, script, params=None): 
        """
        Executes a Gremlin script and returns the Response.

        :param script: Gremlin script to execute.
        :type script: str

        :param params: Param bindings for the script.
        :type params: dict

        :rtype: RexsterResponse

        """
        params = dict(script=script,params=params)
        return self.request.post(gremlin_path,params)


    # Vertex Proxy

    def create_vertex(self, data):
        """
        Creates a vertex and returns the Response.

        :param data: Property data.
        :type data: dict

        :rtype: RexsterResponse

        """
        data = self._remove_null_values(data)
        return self.request.post(vertex_path,data)

    def get_vertex(self, _id):
        """
        Gets the vertex with the _id and returns the Response.

        :param data: Vertex ID.
        :type data: int

        :rtype: RexsterResponse

        """
        path = build_path(vertex_path,_id)
        return self.request.get(path,params=None)

    def get_all_vertices(self):
        """
        Returns a Response containing all the vertices in the Graph.

        :rtype: RexsterResponse

        """
        script = self.scripts.get("get_vertices")
        params = None
        return self.gremlin(script, params)

    def update_vertex(self, _id, data):
        """
        Updates the vertex with the _id and returns the Response.

        :param _id: Vertex ID.
        :type _id: dict

        :param data: Property data.
        :type data: dict

        :rtype: RexsterResponse

        """
        data = self._remove_null_values(data)
        path = build_path(vertex_path,_id)
        return self.request.put(path,data)
        
    def delete_vertex(self, _id):
        """
        Deletes a vertex with the _id and returns the Response.

        :param _id: Vertex ID.
        :type _id: dict

        :rtype: RexsterResponse

        """
        path = build_path(vertex_path,_id)
        return self.request.delete(path,params=None)

    # Edge Proxy

    def create_edge(self, outV, label, inV, data={}): 
        """
        Creates a edge and returns the Response.
        
        :param outV: Outgoing vertex ID.
        :type outV: int

        :param label: Edge label.
        :type label: str

        :param inV: Incoming vertex ID.
        :type inV: int

        :param data: Property data.
        :type data: dict or None

        :rtype: RexsterResponse

        """
        data = self._remove_null_values(data)
        edge_data = dict(_outV=outV,_label=label,_inV=inV)
        data.update(edge_data)
        return self.request.post(edge_path, data)

    def get_edge(self, _id):
        """
        Gets the edge with the _id and returns the Response.

        :param data: Edge ID.
        :type data: int

        :rtype: RexsterResponse

        """
        path = build_path(edge_path, _id)
        return self.request.get(path, params=None)

    def get_all_edges(self):
        """
        Returns a Response containing all the edges in the Graph.

        :rtype: RexsterResponse

        """
        script = self.scripts.get("get_edges")
        params = None
        return self.gremlin(script, params)

    def update_edge(self,_id,data):
        """
        Updates the edge with the _id and returns the Response.

        :param _id: Edge ID.
        :type _id: dict

        :param data: Property data.
        :type data: dict

        :rtype: RexsterResponse

        """
        data = self._remove_null_values(data)
        path = build_path(edge_path, _id)
        return self.request.put(path, data)

    def delete_edge(self,_id):
        """
        Deletes a edge with the _id and returns the Response.

        :param _id: Edge ID.
        :type _id: dict

        :rtype: RexsterResponse

        """
        path = build_path(edge_path, _id)
        return self.request.delete(path, params=None)

    # Vertex Container

    def outE(self,_id, label=None, start=None, limit=None):
        """
        Returns the outgoing edges of the vertex.

        :param _id: Vertex ID.
        :type _id: dict

        :param label: Optional edge label. Defaults to None.
        :type label: str

        :rtype: RexsterResponse
        
        """
        script = self.scripts.get('outE')
        params = dict(_id=_id,label=label,start=start,limit=limit)
        return self.gremlin(script,params)

    def inE(self,_id, label=None, start=None, limit=None):
        """
        Returns the incoming edges of the vertex.

        :param _id: Vertex ID.
        :type _id: dict

        :param label: Optional edge label. Defaults to None.
        :type label: str

        :rtype: RexsterResponse

        """
        script = self.scripts.get('inE')
        params = dict(_id=_id,label=label,start=start,limit=limit)
        return self.gremlin(script,params)

    def bothE(self,_id, label=None, start=None, limit=None):
        """
        Returns the incoming and outgoing edges of the vertex.

        :param _id: Vertex ID.
        :type _id: dict

        :param label: Optional edge label. Defaults to None.
        :type label: str

        :rtype: RexsterResponse
        
        """
        script = self.scripts.get('bothE')
        params = dict(_id=_id,label=label,start=start,limit=limit)
        return self.gremlin(script,params)

    def outV(self,_id, label=None, start=None, limit=None):
        """
        Returns the out-adjacent vertices of the vertex.

        :param _id: Vertex ID.
        :type _id: dict

        :param label: Optional edge label. Defaults to None.
        :type label: str

        :rtype: RexsterResponse

        """
        script = self.scripts.get('outV')
        params = dict(_id=_id,label=label,start=start,limit=limit)
        return self.gremlin(script,params)
        
    def inV(self,_id, label=None, start=None, limit=None):
        """
        Returns the in-adjacent vertices of the vertex.

        :param _id: Vertex ID.
        :type _id: dict

        :param label: Optional edge label. Defaults to None.
        :type label: str

        :rtype: RexsterResponse

        """
        script = self.scripts.get('inV')
        params = dict(_id=_id,label=label,start=start,limit=limit)
        return self.gremlin(script,params)
        
    def bothV(self,_id, label=None, start=None, limit=None):
        """
        Returns the incoming- and outgoing-adjacent vertices of the vertex.

        :param _id: Vertex ID.
        :type _id: dict

        :param label: Optional edge label. Defaults to None.
        :type label: str

        :rtype: RexsterResponse

        """
        script = self.scripts.get('bothV')
        params = dict(_id=_id,label=label,start=start,limit=limit)
        return self.gremlin(script,params)

    # Index Proxy - General

    def get_all_indices(self):
        """Returns a list of all the element indices."""
        return self.request.get(index_path,params=None)

    def get_index(self, name):
        path = build_path(index_path,name)
        return self.request.get(path,params=None)

    def delete_index(self, name): 
        """Deletes the index with the index_name."""
        path = build_path(index_path,name)
        return self.request.delete(path,params=None)
            
    # Index Proxy - Vertex

    def create_vertex_index(self, index_name, *args, **kwds):
        """
        Creates a vertex index with the specified params.

        :param index_name: Name of the index to create.
        :type index_name: str

        :rtype: RexsterResponse

        """
        path = build_path(index_path,index_name)
        index_type = kwds.get('index_type','manual')
        index_keys = kwds.get('index_keys',None)                              
        params = {'class':'vertex','type':index_type}
        if index_keys: 
            params.update({'keys':index_keys})
        return self.request.post(path,params)

    def get_vertex_index(self, index_name):
        """
        Returns the vertex index with the index_name.

        :param index_name: Name of the index.
        :type index_name: str

        :rtype: RexsterResponse

        """
        return self.get_index(index_name)

    def get_or_create_vertex_index(self, index_name, index_params=None):
        script = self.scripts.get('get_or_create_vertex_index')
        params = dict(index_name=index_name, index_params=index_params)
        resp = self.gremlin(script, params)
        #assert "MANUAL" in resp.content['results'][0]
        result = {'name': index_name, 'type': 'manual', 'class': 'vertex'}
        resp.results = RexsterResult(result, self.config)
        return resp

    def delete_vertex_index(self, name): 
        """
        Deletes the vertex index with the index_name.

        :param index_name: Name of the index.
        :type index_name: str

        :rtype: RexsterResponse

        """
        return self.delete_index(name)

    # Index Proxy - Edge

    def create_edge_index(self, name, *args, **kwds):
        """
        Creates a edge index with the specified params.

        :param index_name: Name of the index.
        :type index_name: str

        :rtype: RexsterResponse

        """
        path = build_path(index_path,name)
        index_type = kwds.get('index_type','manual')
        index_keys = kwds.get('index_keys',None)                              
        params = {'class':'edge','type':index_type}
        if index_keys: 
            params.update({'keys':index_keys})
        return self.request.post(path,params)
        
    def get_edge_index(self, name):
        """
        Returns the edge index with the index_name.

        :param index_name: Name of the index.
        :type index_name: str

        :rtype: RexsterResponse

        """
        return self.get_index(name)
        
    def get_or_create_edge_index(self, index_name, index_params=None):
        script = self.scripts.get('get_or_create_edge_index')
        params = dict(index_name=index_name, index_params=index_params)
        resp = self.gremlin(script, params)
        #assert "MANUAL" in resp.content['results'][0]
        result = {'name': index_name, 'type': 'manual', 'class': 'edge'}
        resp.results = RexsterResult(result, self.config)
        return resp

    def delete_edge_index(self, name):
        """
        Deletes the edge index with the index_name.

        :param index_name: Name of the index.
        :type index_name: str

        :rtype: RexsterResponse

        """
        self.delete_index(name)

    #def create_automatic_vertex_index(self,index_name,element_class,keys=None):
    #    keys = json.dumps(keys) if keys else "null"
    #    params = dict(index_name=index_name,element_class=element_class,keys=keys)
    #    script = self.scripts.get('create_automatic_vertex_index',params)
    #    return self.gremlin(script)
        
    #def create_indexed_vertex_automatic(self,data,index_name):
    #    data = json.dumps(data)
    #    params = dict(data=data,index_name=index_name)
    #    script = self.scripts.get('create_automatic_indexed_vertex',params)
    #    return self.gremlin(script)

    # Index Container - General

    def index_count(self, index_name, key, value):
        path = build_path(index_path,index_name,"count")
        params = dict(key=key,value=value)
        return self.request.get(path,params)

    def index_keys(self, index_name):
        path = build_path(index_path,index_name,"keys")
        return self.request.get(path,params=None)

    # Index Container - Vertex

    def put_vertex(self, index_name, key, value, _id):
        """
        Adds a vertex to the index with the index_name.

        :param index_name: Name of the index.
        :type index_name: str

        :param key: Name of the key.
        :type key: str

        :param value: Value of the key.
        :type value: str

        :param _id: Vertex ID
        :type _id: int
        
        :rtype: RexsterResponse

        """
        # Rexster's API only supports string lookups so convert value to a string 
        path = build_path(index_path,index_name)
        params = {'key':key,'value':str(value),'class':'vertex','id':_id}
        return self.request.put(path,params)

    def lookup_vertex(self, index_index_name, key, value):
        """
        Returns the vertices indexed with the key and value.

        :param index_name: Name of the index.
        :type index_name: str

        :param key: Name of the key.
        :type key: str

        :param value: Value of the key.
        :type value: str

        :rtype: RexsterResponse

        """
        path = build_path(index_path,index_index_name)
        params = dict(key=key,value=value)
        return self.request.get(path,params)

    def query_vertex(self, index_name, params):
        """Queries for an edge in the index and returns the Response."""
        path = build_path(index_path,index_name)
        return self.request.get(path,params)

    def remove_vertex(self,index_name,_id,key=None,value=None):
        """
        Removes a vertex from the index and returns the Response.

        :param index_name: Name of the index.
        :type index_name: str

        :param key: Optional. Name of the key.
        :type key: str

        :param value: Optional. Value of the key.
        :type value: str        

        :rtype: RexsterResponse

        """
        # Can Rexster have None for key and value?
        path = build_path(index_path,index_name)
        params = {'key':key,'value':value,'class':'vertex','id':_id}
        return self.request.delete(path,params)

    # Index Container - Edge

    def put_edge(self, index_name, key, value, _id):
        """
        Adds an edge to the index and returns the Response.

        :param index_name: Name of the index.
        :type index_name: str

        :param key: Name of the key.
        :type key: str

        :param value: Value of the key.
        :type value: str

        :param _id: Edge ID
        :type _id: int
        
        :rtype: RexsterResponse

        """
        # Rexster's API only supports string lookups so convert value to a string 
        path = build_path(index_path,index_name)
        params = {'key':key,'value':str(value),'class':'edge','id':_id}
        return self.request.put(path,params)

    def lookup_edge(self, index_index_name, key, value):
        """
        Looks up an edge in the index and returns the Response.

        :param index_name: Name of the index.
        :type index_name: str

        :param key: Name of the key.
        :type key: str

        :param value: Value of the key.
        :type value: str

        :rtype: RexsterResponse

        """
        path = build_path(index_path,index_index_name)
        params = dict(key=key,value=value)
        return self.request.get(path,params)

    def query_edge(self, index_name, params):
        """Queries for an edge in the index and returns the Response."""
        path = build_path(index_path,index_name)
        return self.request.get(path,params)

    def remove_edge(self, index_name, _id, key=None, value=None):
        """
        Removes an edge from the index and returns the Response.
        
        :param index_name: Name of the index.
        :type index_name: str

        :param _id: Edge ID
        :type _id: int

        :param key: Optional. Name of the key.
        :type key: str

        :param value: Optional. Value of the key.
        :type value: str        

        :rtype: RexsterResponse

        """
        # Can Rexster have None for key and value?
        path = build_path(index_path,index_name)
        params = {'key':key,'value':value,'class':'edge','id':_id}
        return self.request.delete(path,params)
    
    # Model Proxy - Vertex

    def create_indexed_vertex(self, data, index_name, keys=None):
        """
        Creates a vertex, indexes it, and returns the Response.

        :param data: Property data.
        :type data: dict

        :param index_name: Name of the index.
        :type index_name: str

        :param keys: Property keys to index.
        :type keys: list

        :rtype: RexsterResponse

        """
        data = self._remove_null_values(data)
        params = dict(data=data,index_name=index_name,keys=keys)
        script = self.scripts.get("create_indexed_vertex")
        return self.gremlin(script,params)
    
    def update_indexed_vertex(self, _id, data, index_name, keys=None):
        """
        Updates an indexed vertex and returns the Response.

        :param index_name: Name of the index.
        :type index_name: str

        :param data: Property data.
        :type data: dict

        :param index_name: Name of the index.
        :type index_name: str

        :param keys: Property keys to index.
        :type keys: list

        :rtype: RexsterResponse

        """
        data = self._remove_null_values(data)
        params = dict(_id=_id,data=data,index_name=index_name,keys=keys)
        script = self.scripts.get("update_indexed_vertex")
        return self.gremlin(script,params)

    # Model Proxy - Edge

    def create_indexed_edge(self, outV, label, inV, data, index_name, keys=None):
        """
        Creates a edge, indexes it, and returns the Response.

        :param outV: Outgoing vertex ID.
        :type outV: int

        :param label: Edge label.
        :type label: str

        :param inV: Incoming vertex ID.
        :type inV: int

        :param data: Property data.
        :type data: dict

        :param index_name: Name of the index.
        :type index_name: str

        :param keys: Property keys to index. Defaults to None (indexes all properties).
        :type keys: list

        :rtype: RexsterResponse

        """
        data = self._remove_null_values(data)
        edge_params = dict(outV=outV,label=label,inV=inV,label_var=self.config.label_var)
        params = dict(data=data,index_name=index_name,keys=keys)
        params.update(edge_params)
        script = self.scripts.get("create_indexed_edge")
        return self.gremlin(script,params)
        
    def update_indexed_edge(self, _id, data, index_name, keys=None):
        """
        Updates an indexed edge and returns the Response.

        :param _id: Edge ID.
        :type _id: int

        :param data: Property data.
        :type data: dict

        :param index_name: Name of the index.
        :type index_name: str

        :param keys: Property keys to index. Defaults to None (indexes all properties).
        :type keys: list

        :rtype: RexsterResponse

        """
        data = self._remove_null_values(data)
        params = dict(_id=_id,data=data,index_name=index_name,keys=keys)
        script = self.scripts.get("update_indexed_edge")
        return self.gremlin(script,params)

    # Utils

    def warm_cache(self):
        """Warms the server cache by loading elements into memory."""
        script = self.scripts.get('warm_cache')
        return self.gremlin(script,params=None)

    # Rexster Specific Stuff

    def rebuild_vertex_index(self, index_name):
        params = dict(index_name=index_name)
        script = self.scripts.get('rebuild_vertex_index',params)
        return self.gremlin(script)

    def rebuild_edge_index(self, index_name):
        params = dict(index_name=index_name)
        script = self.scripts.get('rebuild_edge_index',params)
        return self.gremlin(script)


    # TODO: manual/custom index API

    def multi_get_vertices(self, id_list):
        path = build_path(multi_get_path,"vertices")
        idList = self._build_url_list(id_list)
        params = dict(idList=idList)
        return self.request.get(path,params)

    def multi_get_edges(self, id_list):
        path = build_path(multi_get_path,"edges")
        idList = self._build_url_list(id_list)
        params = dict(idList=idList)
        return self.request.get(path,params)

    def _build_url_list(self, items):
        items = [str(item) for item in items]
        url_list = "[%s]" % ",".join(items)
        return url_list

    def execute_transaction(self, transaction):
        params = dict(tx=transaction.actions)
        return self.request.post(self.transction_path,params)

    def _remove_null_values(self, data):
        """Removes null property values because they aren't valid in Neo4j."""
        # using PUTs to overwrite all properties so no need
        # to worry about deleting props that are being set to null.
        clean_data = [(k, data[k]) for k in data if data[k] is not None]  # Python 3
        return dict(clean_data)




# -*- coding: utf-8 -*-
#
# Copyright 2012 James Thornton (http://jamesthornton.com)
# BSD License (see LICENSE for details)
#
"""
Interface for interacting with a graph database through Rexster.

"""
from bulbs.config import Config
from bulbs.gremlin import Gremlin
from bulbs.element import Vertex, Edge
from bulbs.model import Node, Relationship
from bulbs.base.graph import Graph as BaseGraph

# Rexster-specific imports
from .client import RexsterClient, SAIL_URI
from .index import ManualIndex


class Graph(BaseGraph):
    """
    The primary interface to Rexster.

    Instantiates the database :class:`~bulbs.rexster.client.Client` object using 
    the specified Config and sets up proxy objects to the database.

    :param config: Optional. Defaults to the default config.
    :type config: bulbs.config.Config

    :cvar client_class: RexsterClient class.
    :cvar default_index: Default index class.

    :ivar client: RexsterClient object.
    :ivar vertices: VertexProxy object.
    :ivar edges: EdgeProxy object.
    :ivar config: Config object.
    :ivar gremlin: Gremlin object.
    :ivar scripts: GroovyScripts object.
    
    Example:

    >>> from bulbs.rexster import Graph
    >>> g = Graph()
    >>> james = g.vertices.create(name="James")
    >>> julie = g.vertices.create(name="Julie")
    >>> g.edges.create(james, "knows", julie)

    """
    client_class = RexsterClient
    default_index = ManualIndex
    
    def __init__(self, config=None):
        super(Graph, self).__init__(config)

        # Rexster supports Gremlin
        self.gremlin = Gremlin(self.client)
        self.scripts = self.client.scripts    # for convienience 


    def load_graphml(self,uri):
        """
        Loads a GraphML file into the database and returns the response.

        :param uri: URI of the GraphML file to load.
        :type uri: str

        :rtype: RexsterResult

        """
        script = self.client.scripts.get('load_graphml')
        params = dict(uri=uri)
        return self.gremlin.command(script, params)
        
    def get_graphml(self):
        """
        Returns a GraphML file representing the entire database.

        :rtype: RexsterResult

        """
        script = self.client.scripts.get('save_graphml')
        return self.gremlin.command(script, params=None)
        
    def warm_cache(self):
        """
        Warms the server cache by loading elements into memory.

        :rtype: RexsterResult

        """
        script = self.scripts.get('warm_cache')
        return self.gremlin.command(script, params=None)

    def clear(self):
        """
        Deletes all the elements in the graph.

        :rtype: RexsterResult

        .. admonition:: WARNING 

           This will delete all your data!

        """
        script = self.client.scripts.get('clear')
        return self.gremlin.command(script,params=None)


#
# SailGraph is Undocumented/Experimental - Not Current
#

# TODO: Create a SailClient or sail Client methods.

class SailGraph(object):
    """ An interface to for SailGraph. """

    def __init__(self,root_uri=SAIL_URI):
        self.config = Config(root_uri)
        self.client = RexsterClient(self.config)

        # No indices on sail graphs
        self.gremlin = Gremlin(self.client)        

        self.vertices = VertexProxy(Vertex,self.client)
        self.edges = EdgeProxy(Edge,self.client)

    def add_prefix(self,prefix,namespace):
        params = dict(prefix=prefix,namespace=namespace)
        resp = self.client.post(self._base_target(),params)
        return resp

    def get_all_prefixes(self):
        resp = self.client.get(self._base_target(),params=None)
        return resp.results

    def get_prefix(self,prefix):
        target = "%s/%s" % (self._base_target(), prefix)
        resp = self.client.get(target,params=None)
        return resp.results
        
    def remove_prefix(self,prefix):
        target = "%s/%s" % (self._base_target(), prefix)
        resp = self.client.delete(target,params=None)
        return resp

    def load_rdf(self,url):
        """
        Loads an RDF file into the database, and returns the Rexster 
        response object.

        :param url: The URL of the RDF file to load.

        """
        script = "g.loadRDF('%s', 'n-triples')" % url
        params = dict(script=script)
        resp = self.client.get(self.base_target,params)
        return resp

    def _base_target(self):
        "Returns the base target URL path for vertices on Rexster."""
        base_target = "%s/%s" % (self.client.db_name,"prefixes")
        return base_target
